/*
	alliance.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <stdio.h>
#include "database.h"
#include "calc.h"
#include "error.h"
#include "constants.h"
#include "planet.h"
#include "equip.h"

class alliance //Alliance type
{
	public:
	static const short LIBSIZE=8; //Size of alliance library
	enum {COWARD,ATTACK,DEFENCE,MAKEFRIEND,DESERT,TRAVELHOME,TRAVELFRIEND,TRAVELENEMY,TORPEDO,ATTACKWILD,ATTACKSTRAIGHT,ATTACKFORMATION}; //Different behaviour actions
	enum {TRADE_NONE=0,TRADE_OPEN=1,TRADE_FRIENDS=2,TRADE_CLOSED=3}; //Trading standards constants

	static void init(); //Initialise the datastructures
	static void loadlib(); //Load the alliances from the library in the database
	static void purgeall(); //Purges the alliance data clean
	static alliance* get(short indx); //Returns alliance of given index
	static void maketerritories(); //Generate the territories for all alliances

	bool opposes(alliance* all); //Find out if this alliance opposes the given alliance
	equip* getequip(); //Get a random equipment item that this alliance stocks

	short self; //Self index
	char nam[65]; //Name
	short bhvs[32]; //Behaviour probabilities
	short sinv[16]; //Ship type inventory for this alliance
	short ripo; //Ripoff percentage for purchases (100 is normal)
	short trad; //Trading standards (see enum at top)

	private:
	alliance(short self); //Constructor, giving the alliance its own index in the database
	~alliance(); //Destructor

	void maketerritory(cord seed); //Generate the star systems for this alliance at given seed co-ordinates
	void load(); //Load alliance from database

	short grp; //Ally group
	short nsys; //Number of star systems
	short pinh; //Percentage of inhabited planets
	char cmst[32][65]; //Comm strings
	equip* eqps[16]; //Available equipment	

	static alliance* alliances[LIBSIZE]; //Different available alliances
};

void alliance::init()
{
	for(short i=0;i<LIBSIZE;i++)
		alliances[i]=NULL;
}

void alliance::loadlib()
{
	char nam[12]; //Object name to get

	for(short i=0;i<LIBSIZE;i++)
	{
		sprintf(nam,"Alliance%hd",i);
		try
		{
			database::switchobj(nam);
			alliances[i]=new alliance(i);
			alliances[i]->load();
		}
		catch(error it)
		{
		}
	}
}

void alliance::purgeall()
{
	for(short i=0;i<LIBSIZE;i++)
		if(alliances[i])
			delete alliances[i];
}

alliance* alliance::get(short indx)
{
	if(indx>=0 && indx<LIBSIZE)
	{
		if(alliances[indx])
			return alliances[indx];
		else
			return NULL;
	}
	else
		return NULL;
}

void alliance::maketerritories()
{
	cord seed; //Where to seed each territory from

	for(short i=0;i<LIBSIZE;i++)
	{
		if(alliances[i])
		{
			seed.x=calc::rnd(LIMIT/2)-calc::rnd(LIMIT/2);
			seed.y=calc::rnd(LIMIT/2)-calc::rnd(LIMIT/2);
			alliances[i]->maketerritory(seed);
		}
	}
}

bool alliance::opposes(alliance* all)
{
	if(all==NULL)
		return false;
	if(this->grp!=all->grp)
		return true;
	else
		return false;
}

equip* alliance::getequip()
{
	return eqps[calc::rnd(16)];
}

alliance::alliance(short self) //Constructor, giving the alliance its own index in the database
{
	this->self=self;
}

alliance::~alliance() //Destructor
{
	alliances[self]=NULL;
}

void alliance::maketerritory(cord seed) //Generate the star systems for this alliance at given seed co-ordinates
{
	cord lsys; //Star system location
	char snam[65]; //Name of system
	char pnam[65]; //Name of planet
	short ptyp; //Type of planet

	for(short i=0;i<nsys;i++)
	{
		lsys=seed;
		lsys.x=seed.x+calc::rnd(LIMIT/5)-calc::rnd(LIMIT/5);
		lsys.y=seed.y+calc::rnd(LIMIT/5)-calc::rnd(LIMIT/5);
		for(short j=0,n=calc::rnd(14)+2;j<n;j++)
		{
			if(j==0)
			{
				ptyp=planet::STAR;
				planet::generatename(snam);
				sprintf(pnam,"%s",snam);
			}
			else
			{
				if(calc::rnd(100)<pinh)
				{
					ptyp=planet::INHABITED;
					planet::generatename(pnam);
				}
				else
				{
					ptyp=planet::UNINHABITED;
					sprintf(pnam,"%s %hd",snam,j);
				}
			}
			try
			{
				new planet(pnam,lsys,ptyp,this);
			}
			catch(error it)
			{
			}
			lsys.x+=calc::rnd(10000)-5000;
			lsys.y+=calc::rnd(10000)-5000;
		}
	}
}

void alliance::load()
{
	char atsc[33]; //Attribute scratchpad

	database::getvalue("Name",nam);
	grp=database::getvalue("Group");
	nsys=database::getvalue("Systems");
	pinh=database::getvalue("PercentInhabited");
	ripo=database::getvalue("Ripoff");
	trad=database::getvalue("Trading");
	for(short i=0;i<16;i++)
	{
		sprintf(atsc,"Ship%hd",i);
		sinv[i]=database::getvalue(atsc);
	}
	for(short i=0;i<16;i++)
	{
		sprintf(atsc,"Equipment%hd",i);
		eqps[i]=equip::get(database::getvalue(atsc));
	}
	for(short i=0;i<32;i++)
	{
		sprintf(atsc,"CommString%hd",i);
		database::getvalue(atsc,cmst[i]);
	}
	for(short i=0;i<32;i++)
		bhvs[i]=0;
	bhvs[COWARD]=database::getvalue("Coward");
	bhvs[ATTACK]=database::getvalue("Attack");
	bhvs[MAKEFRIEND]=database::getvalue("MakeFriend");
	bhvs[DESERT]=database::getvalue("Desert");
	bhvs[TRAVELHOME]=database::getvalue("TravelHome");
	bhvs[TRAVELFRIEND]=database::getvalue("TravelFriend");
	bhvs[TRAVELENEMY]=database::getvalue("TravelEnemy");
	bhvs[TORPEDO]=database::getvalue("Torpedo");
	bhvs[ATTACKWILD]=database::getvalue("AttackWild");
	bhvs[ATTACKSTRAIGHT]=database::getvalue("AttackStraight");
	bhvs[ATTACKFORMATION]=database::getvalue("AttackFormation");
}

alliance* alliance::alliances[LIBSIZE];
