/*
	calc.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <math.h>
#include <stdlib.h>
#include <SDL_types.h>
#include <SDL_endian.h>
#include <time.h>

struct vect //A vector
{
	long xx;
	long yy; //X and Y components
};

struct cord //A co-ordinate in the game
{
	long x;
	long y; //X and Y components
};

struct pol //A polar vector
{
        short ang; //Angle
	long rad; //Radius
};

struct box //A universe bounding box
{
	long x1,y1; //Top-left corner
	long x2,y2; //Top-right corner
};

class calc //Mathematics module
{
	public:
	static const double mypi=3.141592654; //Not every OS defines Pi

	static void init(); //Initialise some tables
	static vect poltovect(pol in); //Take a polar, output it as a vector
	static pol vecttopol(vect in); //Take a vector, output it as a polar
	static long quickroot(long x); //Fast integer square root
	static long rnd(long max); //Return random integer from 0 to max-1
	static void getspeed(long spd,char* put); //Convert given game velocity to a string
	static long dattolong(unsigned char* in); //Converts a four byte buffer portably into a long
	static void longtodat(long in,unsigned char* out); //Puts a long portably into a four byte buffer
	static short dattoshort(unsigned char* in); //Converts a two byte buffer portably into a short
	static void shorttodat(short in,unsigned char* out); //Puts a long portably into a four byte buffer
	static bool dateq(unsigned char* d1,unsigned char* d2,short n); //Test two data streams for equality, up to n bytes

	private:
	static long cosc[360]; //Cached cosines
	static long sinc[360]; //Cached sines. Both not actually correct, warped for my preferred co-ordinate system (Up is 0 degrees, go round clockwise, down is Y positive)
	static short angc[100][100]; //Angle cache for vector to polar conversion
	static long wrp[10]; //Warp speed table
	static char spds[33]; //Speed string (saves having to malloc, but it ain't threadsafe!)
};

void calc::init()
{
	double rad; //Angle in radians

	srand(time(NULL));
	for(short i=0;i<360;i++)
	{
		rad=((double)(90-i)/180)*mypi;	
		cosc[i]=(long)(cos(rad)*100);
		sinc[i]=-(long)(sin(rad)*100);
	}
	for(short x=1;x<100;x++)
	{
		for(short y=1;y<100;y++)
		{
			rad=90+(atan((double)y/(double)x)/(mypi/180));
			angc[x][y]=(short)rad;
		}
	}
	for(short i=0;i<10;i++)
	{
		wrp[i]=i*i*i*100;
	}
}

vect calc::poltovect(pol in)
{
	vect out; //Calculated vector

	in.ang=(in.ang%360);
	if(in.ang<0)
		in.ang+=360;
	out.xx=(cosc[in.ang]*in.rad);
	out.yy=(sinc[in.ang]*in.rad);
	out.xx/=100;
	out.yy/=100;
	return out;
}

pol calc::vecttopol(vect in)
{
	pol out; //Calculated polar
	vect ain; //Abs version of in
	
	if(in.xx<0)
		ain.xx=-in.xx;
	else
		ain.xx=in.xx;
	if(in.yy<0)
		ain.yy=-in.yy;
	else
		ain.yy=in.yy;
	if(ain.xx+ain.yy>32000)
	{
		out.rad=ain.xx+ain.yy;
		in.xx/=(1+(out.rad/100));
		in.yy/=(1+(out.rad/100));
	}
	else
	{
		out.rad=quickroot(ain.xx*ain.xx+ain.yy*ain.yy);
		if(out.rad==0)
			out.rad=1;
		in.xx*=100;
		in.yy*=100;
		in.xx/=out.rad;
		in.yy/=out.rad;
	}
	if(in.xx>=100)
		in.xx=99;
	if(in.xx<=-100)
		in.xx=-99;
	if(in.yy>=100)
		in.yy=99;
	if(in.yy<=-100)
		in.yy=-99;
	if(in.yy>0)
	{
		if(in.xx>0)
			out.ang=angc[in.xx][in.yy];
		if(in.xx==0)
			out.ang=180;
		if(in.xx<0)
			out.ang=360-angc[-in.xx][in.yy];
	}
	if(in.yy==0)
	{
		if(in.xx>0)
			out.ang=90;
		else
			out.ang=270;
	}
	if(in.yy<0)
	{
		if(in.xx>0)
			out.ang=180-angc[in.xx][-in.yy];
		if(in.xx==0)
			out.ang=0;
		if(in.xx<0)
			out.ang=180+angc[-in.xx][-in.yy];
	}
//	if(out.rad==0)
//		out.rad=1;
	if(in.xx==0 && in.yy==0)
		out.rad=0;
	return out;
}

long calc::quickroot(long x)
{
	long y1,y2; //Iteration variables
	bool ascl; //Scaling used for accuracy?

	if(x<1)
		return 0;
	if(x<10000)
		ascl=true;
	else
		ascl=false;
	if(ascl)
		x*=900;
	y1=x/2;
	for(short i=0;i<15;i++)
	{
		if(y1==0)
			y1=1;
		y2=(y1+x/y1)/2;
		y1=y2;
	}
	if(ascl)
		y2/=30;
	return y2;
}

long calc::rnd(long max)
{
	if(max==0)
		return 0;
	return (rand()/((RAND_MAX/max)+1));
}

void calc::getspeed(long spd,char* put)
{
	short j; //To extract the value from the loop

	j=0;
	for(short i=9;i>=0;i--)
	{
		if(spd>=wrp[i])
		{
			j=i;
			break;
		}
	}
	if(j==0)
		if(spd<10)
			sprintf(put,"0.0%ld c",spd);
		else
			sprintf(put,"0.%ld c",spd);
	else
		sprintf(put,"Warp %hd",j);
}

long calc::dattolong(unsigned char* in)
{
	Uint32 tmp; //Temporary value holder
	long out; //Value to output

	tmp=*((Uint32*)in);
	#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	tmp=SDL_Swap32(tmp);
	#endif	
	out=tmp-2147483647;

	return out;
}

void calc::longtodat(long in,unsigned char* out)
{
	Uint32 tmp; //Temporary value holder

	tmp=in+2147483647;
	#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	tmp=SDL_Swap32(tmp);
	#endif	
	*((Uint32*)out)=tmp;
	return;
}

short calc::dattoshort(unsigned char* in)
{
	Uint16 tmp; //Temporary value holder
	short out; //Value to output

	tmp=*((Uint16*)in);
	#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	tmp=SDL_Swap16(tmp);
	#endif	
	out=tmp-32768;

	return out;
}

void calc::shorttodat(short in,unsigned char* out)
{
	Uint16 tmp; //Temporary value holder

	tmp=in+32768;
	#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	tmp=SDL_Swap16(tmp);
	#endif	
	*((Uint16*)out)=tmp;
	return;
}

bool calc::dateq(unsigned char* d1,unsigned char* d2,short n)
{
	for(short i=0;i<n;i++)
		if(d1[i]!=d2[i])
			return false;
	return true;		
}

long calc::cosc[360];
long calc::sinc[360];
short calc::angc[100][100];
long calc::wrp[10];
char calc::spds[33];
