/*
	camera.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <stdio.h>
#include "calc.h"
#include "graphic.h"
#include "interface.h"
#include "protocol.h"
#include "presence.h"
#include "constants.h"
#include "sound.h"
#include "error.h"

#include "ship.h"
#include "planet.h"


#define x2screen(rx) (long)cx+rx-pov.x
#define y2screen(ry) (long)cy+ry-pov.y

struct star //Background star
{
	cord loc; //Location
	short dep; //Depth
	short rot; //Rotation (for warp stars)
};

class camera //Viewer of the client's universe
{
	public:
	static void init(); //Initialise camera datastructures
	static void turnon(); //Turn camera on
	static void turnoff(); //Turn camera off
	static void bind(presence* src); //Bind the camera to given source
	static void unbind(); //Camera floats free, no longer bound
	static void render(); //Render what the camera can see
	static void update(); //Update camera position
	static void noise(sound* snd,presence* src); //Play noise from given source
	static void shake(short mag); //Shake the camera with given magnitude
	static void radarzoom(short dir); //Zoom the sensor in or out, with dir +1 or -1 respectively

	static box cov; //Total camera coverage

	private:
	static void rendermainview(); //Render the main screen view
	static void renderstars(); //Render the background stars
	static void renderradar(); //Render the radar view
	
	static bool on; //Is camera turned on?
	static presence* src; //Camera source
	static cord pov; //Camera point of view in universe
	static long zoom; //Sensor zoom out
	static star strs[64]; //Background stars
	static short shak; //Current shaking magnitude
	static bool flck; //Flicker for radar
};

void camera::init()
{
	src=NULL;
	on=false;
	for(short i=0;i<64;i++)
	{
		strs[i].loc.x=2*LIMIT;
		strs[i].loc.y=2*LIMIT;
		strs[i].dep=1;
	}
}

void camera::turnon()
{
	on=true;
}

void camera::turnoff()
{
	on=false;
}

void camera::bind(presence* src)
{
	if(camera::src!=src)
	{
		camera::src=src;
		zoom=presence::srng;
		update();
	}
}

void camera::unbind()
{
	src=NULL;
}

void camera::render()
{
	if(!on)
		return;

	graphic::clip(&interface::viewb);
	graphic::box(&interface::viewb,graphic::BLACK);
	renderstars();
	rendermainview();

	graphic::clip(&interface::radarb);
	graphic::box(&interface::radarb,graphic::BLACK);
	if(src)
		renderradar();
}

void camera::update()
{
	cord opov; //Old pov (for interpolation

	opov=pov;
	if(src)
		pov=src->loc;
	if(shak)
	{
		shak--;
		pov.x+=calc::rnd(shak*2)-shak;
		pov.y+=calc::rnd(shak*2)-shak;
	}
	cov.x1=pov.x-zoom;
	cov.x2=pov.x+zoom;
	cov.y1=pov.y-zoom;
	cov.y2=pov.y+zoom;
}

void camera::noise(sound* snd,presence* src)
{
	vect ssrc; //Sound source vector
	long dis; //Rough distance from PoV

	ssrc.xx=src->loc.x-pov.x;
	ssrc.yy=src->loc.y-pov.y;
	if(ssrc.xx<0)
		ssrc.xx=-ssrc.xx;
	if(ssrc.yy<0)
		ssrc.yy=-ssrc.yy;
	dis=ssrc.xx+ssrc.yy;
	if(dis<600)
	{
		snd->play(1);
		return;
	}
	if(dis<6400)
	{
		snd->play(3);
		return;
	}
	if(dis<12800)
	{
		snd->play(6);
		return;
	}	
}

void camera::shake(short mag)
{
	if(mag>20)
		mag=20;
	shak=mag;
}

void camera::radarzoom(short dir)
{
	if(dir==+1)
		if(zoom>0)
			zoom=(zoom*2)/3;
	if(dir==-1)
		if(zoom>0)
			zoom=(zoom*3)/2;
	if(zoom>presence::lrng)
		zoom=presence::lrng;
	if(zoom<200)
		zoom=200;
}

box camera::cov;

void camera::rendermainview()
{
	graphic* ptr; //Target pointer sprite
	char txt[33]; //For rendering distance on the pointer
	presence* tprs; //Pointer to objects to draw
	short cx,cy; //Centering screen position
	long sx,sy; //Screen co-ordinates
	pol pptr;
	vect vptr; //Pointer to target

	cx=interface::viewb.x+(interface::viewb.w/2);
	cy=interface::viewb.y+(interface::viewb.h/2);
	for(short i=PT_PLANET;i<=PT_FRAG;i++)
	{
		for(short j=0;j<presence::ISIZE;j++)
		{
			tprs=presence::get(j);
			if(tprs && tprs->typ==i)
			{
				sx=x2screen(tprs->loc.x);
				sy=y2screen(tprs->loc.y);
				if(sx>interface::viewb.x-100 && sx<interface::viewb.x+interface::viewb.w+100 && sy>interface::viewb.y-100 && sy<interface::viewb.y+interface::viewb.h+100)
					tprs->drawat(sx,sy);
			}
		}
	}

	if(presence::trg)
	{
		vptr.xx=presence::trg->loc.x-pov.x;
		vptr.yy=presence::trg->loc.y-pov.y;
		pptr=calc::vecttopol(vptr);
		sprintf(txt,"%ld",(long)pptr.rad/100);
		pptr.rad=interface::viewb.w/2-50;
		vptr=calc::poltovect(pptr);
		vptr.xx+=interface::viewb.x+interface::viewb.w/2;
		vptr.yy+=interface::viewb.y+interface::viewb.h/2;
		ptr=graphic::get(graphic::NAV);
		if(ptr)
			ptr->draw(vptr.xx,vptr.yy,pptr.ang/10,0,0,false);
		graphic::string(txt,vptr.xx,vptr.yy+4);
	}
}

void camera::renderstars()
{
	short cx,cy; //Centering screen position
	short astx,asty; //For holding star position on screen
	graphic* warp; //Warp star sprite
	
	warp=NULL;
	cx=interface::viewb.x+(interface::viewb.w/2);
	cy=interface::viewb.y+(interface::viewb.h/2);
	if(presence::vel.rad>=99)
		warp=graphic::get(graphic::WARP);
	for(short i=0;i<64;i++)
	{
		if(warp && strs[i].dep>10)
			strs[i].dep=1;
		astx=(short)(strs[i].loc.x+(cx-pov.x)/strs[i].dep);
		asty=(short)(strs[i].loc.y+(cy-pov.y)/strs[i].dep);
		if(astx<interface::viewb.x || astx>interface::viewb.x+interface::viewb.w || asty<interface::viewb.y || asty>interface::viewb.y+interface::viewb.h)
		{
			astx=interface::viewb.x+calc::rnd(interface::viewb.w);
			asty=interface::viewb.y+calc::rnd(interface::viewb.h);

			if(warp)
				strs[i].dep=calc::rnd(40)+60;
			else
				strs[i].dep=calc::rnd(10)+1;

			strs[i].loc.x=astx-(interface::viewb.x+interface::viewb.w/2-pov.x)/strs[i].dep;
			strs[i].loc.y=asty-(interface::viewb.y+interface::viewb.h/2-pov.y)/strs[i].dep;
		}
		if(warp)
			warp->draw(astx,asty,((presence::vel.ang+5)/10)%36,0,0,false);
		else
			if(strs[i].dep>5)
				graphic::pix(astx,asty,graphic::GREY);
			else
				graphic::pix(astx,asty,graphic::WHITE);
	}
}

void camera::renderradar()
{
	long sx,sy; //Screen co-ordinates
	presence* tprs; //Pointer to objects to draw
	char txt[33]; //For rendering co-ordinates on radar
	short col; //Color to use on radar
	sbox tbox; //Target drawing box

	flck=!flck;
	if((cov.x2-cov.x1)<(presence::srng*4))
	{
		for(long x=(long)cov.x1/2000,l=(long)cov.x2/2000;x<=l;x++)
		{
			tbox.x=(short)((((x*2000)-pov.x)*(interface::radarb.w/2))/(cov.x2-pov.x)+interface::radarb.x+interface::radarb.w/2);
			tbox.y=interface::radarb.y+interface::radarb.h-2;
			tbox.w=1;
			tbox.h=3;
			graphic::box(&tbox,graphic::GREY);
		}
		for(long y=(long)cov.y1/2000,l=(long)cov.y2/2000;y<=l;y++)
		{
			tbox.y=(short)((((y*2000)-pov.y)*(interface::radarb.h/2))/(cov.y2-pov.y)+interface::radarb.y+interface::radarb.h/2);
			tbox.x=interface::radarb.x+interface::radarb.w-2;
			tbox.w=3;
			tbox.h=1;
			graphic::box(&tbox,graphic::GREY);
		}
	}

	if(cov.x2-pov.x>presence::srng)
	{
		tbox.x=(-presence::srng)/((cov.x2-cov.x1)/(long)interface::radarb.w)+interface::radarb.x+interface::radarb.w/2;
		tbox.y=(-presence::srng)/((cov.y2-cov.y1)/(long)interface::radarb.h)+interface::radarb.y+interface::radarb.h/2;
		tbox.w=(presence::srng)/((cov.x2-cov.x1)/(long)interface::radarb.w)+interface::radarb.x+interface::radarb.w/2;
		tbox.h=(presence::srng)/((cov.y2-cov.y1)/(long)interface::radarb.h)+interface::radarb.y+interface::radarb.h/2;
		graphic::line(tbox.x,tbox.y,tbox.w,tbox.y,graphic::LIGHTBLUE);
		graphic::line(tbox.w,tbox.y,tbox.w,tbox.h,graphic::LIGHTBLUE);
		graphic::line(tbox.w,tbox.h,tbox.x,tbox.h,graphic::LIGHTBLUE);
		graphic::line(tbox.x,tbox.h,tbox.x,tbox.y,graphic::LIGHTBLUE);
	}

	for(short i=0;i<presence::ISIZE;i++)
	{
		tprs=presence::get(i);
		if(tprs)
		{
			sx=(tprs->loc.x-pov.x)/((cov.x2-cov.x1)/(long)interface::radarb.w)+interface::radarb.x+interface::radarb.w/2;
			sy=(tprs->loc.y-pov.y)/((cov.y2-cov.y1)/(long)interface::radarb.h)+interface::radarb.y+interface::radarb.h/2;
			if(tprs->loc.x>cov.x1 && tprs->loc.x<cov.x2 && tprs->loc.y>cov.y1 && tprs->loc.y<cov.y2)
			{
				switch(tprs->typ)
				{
					case PT_SHIP:
					if(tprs->enem)
						col=graphic::RED;
					else
						col=graphic::WHITE;
					break;
	
					case PT_PLANET:
					col=graphic::GREEN;
					break;

					case PT_FRAG:
					col=graphic::BLUE;
					break;
				}
				if(tprs==presence::me)
					if(flck)
						col=graphic::RED;
					else
						col=graphic::BLUE;
				if(tprs==presence::trg || tprs==presence::hl)
				{
					tbox.x=(short)sx-1;
					tbox.y=(short)sy-1;
					tbox.w=3;
					tbox.h=3;
					if(tprs==presence::hl)
					{
						if(flck)
							graphic::box(&tbox,col);
					}
					else
						graphic::box(&tbox,col);
				}
				else
					graphic::pix(sx,sy,col);
			}
		}
	}
	if(presence::trg)
	{
		graphic::string(presence::trg->nam,interface::radarb.x+1,interface::radarb.y+interface::radarb.h-13);
		graphic::string(presence::trg->anno,interface::radarb.x+1,interface::radarb.y+interface::radarb.h-6);
	}
	sprintf(txt,"%ld , %ld",(long)pov.x/100,(long)pov.y/100);
	graphic::string(txt,interface::radarb.x,interface::radarb.y);
	calc::getspeed(presence::vel.rad,txt);
	graphic::string(txt,interface::radarb.x+interface::radarb.w-6*strlen(txt)-2,interface::radarb.y);
}

bool camera::on;
presence* camera::src;
cord camera::pov;
long camera::zoom;
star camera::strs[64];
short camera::shak;
bool camera::flck;
