/*
	client.cc	
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <string.h>
#include "calc.h"
#include "sockhelper.h"
#include "protocol.h"
#include "sound.h"
#include "presence.h"
#include "camera.h"
#include "error.h"
#include "interface.h"
#include "graphic.h"

class client //Module to handle connection to servers
{
	public:
	static void init(); //Initialise the client datastructures
	static void stop(); //Clean up and terminate client, disconnecting from the server
	static void connect(char* host); //Connect to given host
	static void poll(); //Run the client for a cycle, polling for new objects and acting on them
	static void action(short typ,long opr); //Send an action to the server, given action number and operand

	private:
	static void readln(); //Handles a server readline request until it's done, with password hiding style optional

	static bool edit; //In edit mode?
	static bool hide; //In passworded edit?
	static TCPsocket sock; //TCP socket to server
	static sockhelper* hlpr; //Socket wrapper
	static short btck; //Byte count ticker
};

void client::init()
{
	sock=NULL;
	hlpr=NULL;
	edit=false;
}

void client::stop()
{
	if(sock)
	{
		SDLNet_TCP_Close(sock);
		sock=NULL;
	}
	if(hlpr)
	{
		delete hlpr;
		hlpr=NULL;
	}
}

void client::connect(char* host)
{
	IPaddress serv; //Server address

	SDLNet_ResolveHost(&serv,host,PORT);
	if(serv.host==INADDR_NONE)
		throw error("Couldn't resolve hostname");
	sock=SDLNet_TCP_Open(&serv);
	if(!sock)
		throw error("Could not connect to server");
	hlpr=new sockhelper(sock);
	hlpr->send((unsigned char*)SIGN,strlen(SIGN));
	btck=0;
}

void client::poll()
{
	unsigned char otyp; //Incoming object type
	unsigned char* buf; //Incoming buffer to use
	sound* snd; //Sound to play
	graphic* cspr; //Console sprite
	short n; //Array subscript to access
	presence* from; //Sound source if applicable
	short len; //Length of console/message text if applicable
	char txt[1025]; //Text to print to console

	hlpr->pump();
	btck=(btck+1)%24;
	if(btck==0)
	{
		long n=hlpr->getcount();
		action(CLIENT_BANDWIDTH,n);
	}
	while(hlpr->request(1))
	{
		otyp=*hlpr->request(1);
		switch(otyp)
		{
			case SERV_SELF:
			buf=hlpr->request(SERV_SELF_SZ);
			if(buf)
			{
				presence::feed(buf);
				hlpr->suck();
			}
			break;

			case SERV_NEW:
			buf=hlpr->request(SERV_NEW_SZ);
			if(buf)
			{
				presence::feed(buf);
				hlpr->suck();
			}
			break;

			case SERV_NAME:
			buf=hlpr->request(SERV_NAME_SZ);
			if(buf)
			{
				presence::feed(buf);
				hlpr->suck();
			}
			break;

			case SERV_UPD:
			buf=hlpr->request(SERV_UPD_SZ);
			if(buf)
			{
				presence::feed(buf);
				hlpr->suck();
			}
			break;

			case SERV_DEL:
			buf=hlpr->request(SERV_DEL_SZ);
			if(buf)
			{
				presence::feed(buf);
				hlpr->suck();
			}
			break;

			case SERV_HILIGHT:
			buf=hlpr->request(SERV_HILIGHT_SZ);
			if(buf)
			{
				presence::feed(buf);
				hlpr->suck();
			}
			break;
				
			case SERV_SND:
			buf=hlpr->request(SERV_SND_SZ);
			if(buf)
			{
				snd=sound::get(calc::dattoshort(buf+1));
				if(snd)
					snd->play(1);
				hlpr->suck();
			}
			break;

			case SERV_NOISE:
			buf=hlpr->request(SERV_NOISE_SZ);
			if(buf)
			{
				snd=sound::get(calc::dattoshort(buf+1));
				n=calc::dattoshort(buf+3);
				if(n>=0 && n<presence::ISIZE)
					from=presence::get(n);
				else
					from=NULL;
				if(snd && from)
					camera::noise(snd,from);
				hlpr->suck();
			}
			break;

			case SERV_SHAKE:
			buf=hlpr->request(SERV_SHAKE_SZ);
			if(buf)
			{
				camera::shake(calc::dattoshort(buf+1));
				hlpr->suck();
			}
			break;

			case SERV_CONS:
			buf=hlpr->request(3);
			if(buf)
			{
				edit=false;
				len=calc::dattoshort(buf+1);
				if(len>1024)
					throw error("Overflow from server in SERV_CONS");
				buf=hlpr->request(3+len);
				if(buf)
				{
					memcpy(txt,buf+3,len);
					txt[len]='\0';
					interface::printtocons("%s\n",txt);
					hlpr->suck();
				}
			}
			break;
			
			case SERV_READLN:
			buf=hlpr->request(SERV_READLN_SZ);
			if(buf)
			{
				edit=true;
				if(buf[1])
					hide=true;
				else
					hide=false;
				hlpr->suck();
			}
			break;

			case SERV_MESG:
			buf=hlpr->request(3);
			if(buf)
			{
				len=calc::dattoshort(buf+1);
				if(len>128)
					throw error("Overflow from server in SERV_MESG");
				buf=hlpr->request(3+len);
				if(buf)
				{
					memcpy(txt,buf+3,len);
					txt[len]='\0';
					interface::printtomesg("%s",txt);
					hlpr->suck();
				}
			}
			break;

			case SERV_CSPR:
			buf=hlpr->request(3);
			if(buf)
			{
				cspr=graphic::get(calc::dattoshort(buf+1));
				if(cspr)
					interface::spritetocons(cspr);
				hlpr->suck();
			}
			break;

			case SERV_FLOOD:
			buf=hlpr->request(SERV_FLOOD_SZ);
			if(buf)
				hlpr->suck();
			break;

			default:
			break;
		}
		hlpr->pump();
		if(!buf)
			return;
	}
	if(edit)
		readln();
}

void client::action(short typ,long opr)
{
	unsigned char buf[3]; //Outgoing buffer

	if(opr>32767)
		opr=32767;
	if(opr<-32766)
		opr=-32766;
	buf[0]=(unsigned char)typ;
	calc::shorttodat(opr,buf+1);
	hlpr->send(buf,3);
}

void client::readln()
{
	char txt[65]; //Readline text

	if(interface::getline(txt,hide))
	{
		edit=false;
		for(short i=0;i<65;i++)
		{
			action(CLIENT_CHAR,txt[i]);
			if(txt[i]=='\0')
				break;
		}
	}
}

bool client::edit;
bool client::hide;
TCPsocket client::sock;
sockhelper* client::hlpr;
short client::btck;
