/*
	database.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <stdio.h>
#include <string.h>
#include "error.h"

struct obj //An object entry in the hash table
{
	char* nam; //Name of object
	FILE* strm; //Filestream containing object
	long pos; //Position in file
	long len; //Length of entire object entry
	obj* next; //Pointer to next element in linked list
	obj* prev; //Pointer to previous element in linked list
};

struct reader //Part of a stack of opened readers
{
	FILE* strm; //Associated filestream
	reader* next; //Next down in stack
};

class database //Object database spanning multiple files
{
	public:
	static void init(); //Initialise and blank the datastructures
	static void openreader(char* path); //Opens a data file and parses it into the database
	static void openwriter(char* path); //Opens a file to write objects to
	static void closereader(); //Closes the current writer
	static void closewriter(); //Closes the current writer
	static void switchobj(char* name); //Jump to the named object in the database, returning if it was found or not
	static char* getvalue(char* key,char* val); //Get the value of the given attribute (space to put value should be specified)
	static long getvalue(char* key); //As above, but value parsed to be an integer
	static void putobject(char* name); //Write the header of an object to the open writer
	static void putvalue(char* key,char* val); //Write key=value to the open writer
	static void putvalue(char* key,long val);

	private:
	static void submitobj(FILE* strm,char* nam,long pos,long len); //Submit an object to the location hash, detailing which file it's in, its name, its position within the file and length within the file (strm is a filestream, dodgy workaround necessitates 'void')
	static obj* locateobj(char* nam); //Locate and return a pointer to the object holder referring to object of given name, NULL if not found
	static short hashstring(char* str); //Hashes the given string to a 0-1023 value

	static obj* bcks[1024]; //Hash table buckets for object entries
	static char* ostr; //Temporary loaded object store
	static reader* ostk; //Stack of open readers
	static FILE* owrt; //Writer filestream
	
	static const char MAGIC[]="SV0040"; //Magic number for database files
};

void database::init()
{
	for(short i=0;i<1024;i++)
		bcks[i]=NULL;
	ostr=NULL;
	ostk=NULL;
	owrt=NULL;
}

void database::openreader(char* path)
{
	long pos; //Current position in file
	short lr; //Length read from file
	char* fnd; //String finding pointer
	FILE* strm; //Open filestream
	char onam[65]; //Object name found
	long opos; //Object position found
	char tmp[2049]; //Temporary reading buffer
	reader* rstk; //Reader to put on the stack

	strm=fopen(path,"r");
	if(!strm)
		throw error("Datafile not found");
	tmp[0]='\0';
	lr=fread(tmp,1,strlen(MAGIC),strm);
	tmp[lr]='\0';
	if(strcmp(tmp,MAGIC)!=0)
		throw error("File is corrupt, or incompatible with this version");
	rstk=new reader;
	rstk->strm=strm;
	rstk->next=NULL;
	if(ostk)
		rstk->next=ostk;
	ostk=rstk;
	lr=2;
	opos=-1;
	while(lr>1)
	{
		pos=ftell(strm);
		lr=fread(tmp,1,256,strm);
		tmp[lr]='\0';
		if(tmp[0]=='\n' && tmp[1]=='@')
		{
			pos++;
			fseek(strm,pos,SEEK_SET);
			lr=fread(tmp,1,66,strm);
			tmp[lr]='\0';
			fnd=strstr(tmp,"\n");
			if(fnd)
			{
				if(opos!=-1)
					submitobj(strm,onam,opos,pos-opos-1);
				*fnd='\0';
				sprintf(onam,"%s",tmp+1);
				opos=pos;
			}
			fseek(strm,pos,SEEK_SET);
		}
		else
		{
			fnd=strstr(tmp,"\n@");
			if(fnd)
				fseek(strm,fnd-tmp+pos,SEEK_SET);
			else
				fseek(strm,-1,SEEK_CUR);
		}
	}
	if(opos!=-1)
		submitobj(strm,onam,opos,ftell(strm)-opos-1);
}

void database::openwriter(char* path) //Open a writer into the file at the given path
{
	closewriter();
	owrt=fopen(path,"w");
	if(!owrt)
		throw error("Error opening writer");
	fprintf(owrt,"%s\n",MAGIC);
}

void database::closereader() //Close the last opened reader
{
	FILE* strm; //Stream to close
	reader* del; //Reader to delete
	obj* curr;
	obj* next; //Objects to inspect

	if(ostk)
	{
		del=ostk;
		strm=del->strm;
		fclose(strm);
		ostk=del->next;
		delete del;
		for(short i=0;i<1024;i++)
		{
			curr=bcks[i];
			while(curr)
			{
				if(curr->strm==strm)
				{
					if(curr->next)
						curr->next->prev=curr->prev;
					next=curr->next;
					if(curr->prev)
						curr->prev->next=curr->next;
					else
						bcks[i]=curr->next;
					delete curr;
					curr=next;
				}
				else
					curr=curr->next;
			}
		}
	}
}

void database::closewriter()
{
	if(owrt)
	{
		fclose(owrt);
		owrt=NULL;
	}
}

void database::switchobj(char* nam)
{
	obj* got; //Object got from database

	got=locateobj(nam);
	if(got)
	{
		if(ostr)
			delete ostr;
		ostr=new char[got->len+1];
		fseek(got->strm,got->pos,SEEK_SET);
		ostr[fread(ostr,1,got->len,got->strm)]='\0';
	}
	else
	{
		if(ostr)
			delete ostr;
		ostr=NULL;
		throw error("Object not found in database");
	}
}

char* database::getvalue(char* key,char* val)
{
	char srch[68]; //Key statement to search for	
	char* fnd; //Pointer to found string
	short lk; //Length of key
	char* out; //Value to return

	out=val;
	lk=strlen(key);
	if(lk>64)
		return NULL;
	if(ostr)
	{
		sprintf(srch,"\n%s =",key);
		fnd=strstr(ostr,srch);
		if(!fnd)
		{
			sprintf(srch,"\n%s=",key);
			fnd=strstr(ostr,srch);
		}
		if(fnd)
		{
			fnd+=strlen(srch);
			while(*fnd==' ')
				fnd++;
			for(short i=0;i<64 && *fnd!='\0' && *fnd!='\n';i++)
			{
				*val=*fnd;
				val++;
				fnd++;
			}
			*val='\0';
		}
		else
			out[0]='\0';
	}
	else
		out[0]='\0';
	return out;
}

long database::getvalue(char* key)
{
	char val[65]; //String representation
	long out; //Value to output

	getvalue(key,val);
	out=-1;
	sscanf(val,"%ld",&out);
	return out;
}

void database::putobject(char* nam)
{
	if(owrt)
		fprintf(owrt,"@%s\n",nam);
}

void database::putvalue(char* key,char* val)
{
	if(owrt)
		fprintf(owrt,"%s=%s\n",key,val);
}

void database::putvalue(char* key,long val)
{
	if(owrt)
		fprintf(owrt,"%s=%ld\n",key,val);
}

void database::submitobj(FILE* strm,char* nam,long pos,long len)
{
		short hash; //Hash of name
	obj* next; //Next bucket in chain to remember

	hash=hashstring(nam);
	if(bcks[hash])
		next=bcks[hash];
	else
		next=NULL;
	bcks[hash]=new obj;
	bcks[hash]->nam=new char[strlen(nam)+1];
	sprintf(bcks[hash]->nam,"%s",nam);
	bcks[hash]->strm=strm;
	bcks[hash]->pos=pos;
	bcks[hash]->len=len;
	bcks[hash]->next=next;
	if(next)
		next->prev=bcks[hash];
	bcks[hash]->prev=NULL;
}

obj* database::locateobj(char* nam)
{
	obj* next; //Next object to try
	short hash; //Hash of name

	hash=hashstring(nam);
	next=bcks[hash];
	while(next)
	{
		if(strcmp(nam,next->nam)==0)
			return next;
		next=next->next;
	}
	return NULL;
}

short database::hashstring(char* str)
{
	long out; //Value to output

	out=1;
	for(short i=0,j=strlen(str);i<j;i++)
		out=(out*str[i])%1024;
	return out;
}

obj* database::bcks[1024];
char* database::ostr;
reader* database::ostk;
FILE* database::owrt;
const char database::MAGIC[];
