/*
	frag.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <signal.h>
#include <SDL.h>
#include "calc.h"
#include "database.h"
#include "error.h"
#include "ship.h"
#include "planet.h"
#include "protocol.h"

class frag //A frag object, like phaser fire, torpedo, debris
{
	public:
	enum {ENERGY=1,HOMER=2,DEBRIS=3}; //Frag types
	static const short FIRE=1; //Index of burning fire graphic
	static const short ISIZE=512; //Capacity of frag index

	frag(cord loc,short typ,short spr,short col,ship* trg,ship* own,vect mov,short rot,short pow,short trck,short rng); //Constructor to create a new frag from scratch

	~frag(); //Destructor

	static void init(); //Initialise the frag system datastructures
	static void purgeall(); //Clean out the frag system after use
	static void simulateall(); //Simulate all the frags and physics
	static void notifydeath(ship* tshp); //Notify the frags of the deletion of a ship, to resolve dangling target and owner references
	static void saveall(); //Save all frags to the database
	static void loadall(); //Load all frags from database
	static frag* get(short indx); //Return the frag of given index

	void netout(short typ,unsigned char* buf); //Get type of data from frag into a network buffer

	short self; //Self index in the database
	cord loc; //Position

	private:
	frag(short self); //Constructor, give it its index value and it will load from the database

	void physics(); //Move the frag, do collisions
	void home(); //Sub-function of physics, handles homing
	void save(); //Save frag to a database
	void load(); //Load frag from a database

	static frag* frags[ISIZE]; //The frag index
	static short sel; //The slot selector

	short typ; //Type (see the top enum)
	short spr; //Associated sprite index
	short col; //Colour if relevant (makes it a beam weapon)
	ship* trg; //Target ship index
	ship* own; //Owner ship index
	long ox,oy; //Old co-ordinates, for purposes of clientside interpolation
	vect mov; //Velocity vector
	short rot; //Rotation frame to use
	short pow; //Power/damage
	long trck; //Tracking power
	short rng; //Range left
};

frag::frag(cord loc,short typ,short spr,short col,ship* trg,ship* own,vect mov,short rot,short pow,short trck,short rng)
{
	if(frags[sel])
		delete frags[sel];
	self=sel;	
	frags[self]=this;
	sel++;
	if(sel>=ISIZE)
		sel=0;

	this->loc=loc;
	this->typ=typ;
	this->spr=spr;
	this->col=col;
	this->trg=trg;
	this->own=own;
	this->mov=mov;
	this->rot=rot;
	this->pow=pow;
	this->trck=trck;
	this->rng=rng;
}

frag::~frag()
{
	if(self!=-1)
		frags[self]=NULL;
}

void frag::init()
{
	for(short i=0;i<ISIZE;i++)
		frags[i]=NULL;
	sel=0;
}

void frag::purgeall()
{
	for(short i=0;i<ISIZE;i++)
		if(frags[i])
			delete frags[i];
}

void frag::simulateall()
{
	for(short i=0;i<ISIZE;i++)
	{
		if(frags[i])
			frags[i]->physics();
		if(frags[i] && frags[i]->trg)
			frags[i]->home();
	}	
}

void frag::notifydeath(ship* tshp)
{
	for(short i=0;i<ISIZE;i++)
	{
		if(frags[i])
		{
			if(frags[i]->trg==tshp)
				frags[i]->trg=NULL;
			if(frags[i]->own==tshp)
				frags[i]->own=NULL;
		}
	}
}

void frag::saveall()
{
	char obsc[33]; //Object name scratchpad

	for(short i=0;i<ISIZE;i++)
	{
		sprintf(obsc,"Frag%hd",i);
		if(frags[i])
		{
			database::putobject(obsc);
			frags[i]->save();
		}
	}
}

void frag::loadall()
{
	for(short i=0;i<ISIZE;i++)
	{
		try
		{
			new frag(i);
		}
		catch(error it)
		{
		}
	}
}

frag* frag::get(short indx)
{
	if(indx>=0 && indx<ISIZE)
		if(frags[indx])
			return frags[indx];
		else
			return NULL;
	else
		return NULL;
}

void frag::netout(short typ,unsigned char* buf)
{
	buf[0]=typ;
	buf+=1;

	calc::shorttodat(frag2pres(self),buf);
	buf+=2;
	switch(typ)
	{
		case SERV_NEW:
		*buf=PT_FRAG;
		buf+=1;
		if(col>=0)
			calc::shorttodat(-col,buf);
		else
			calc::shorttodat(spr,buf);
		buf+=2;
		if(own)
			calc::shorttodat(ship2pres(own->self),buf);
		else
			calc::shorttodat(-1,buf);
		break;

		case SERV_UPD:
		calc::longtodat(loc.x,buf);
		buf+=4;
		calc::longtodat(loc.y,buf);
		buf+=4;
		calc::longtodat(mov.xx,buf);
		buf+=4;
		calc::longtodat(mov.yy,buf);
		buf+=4;
		calc::shorttodat(rot*10,buf);
		buf+=2;
		*buf=0;
		buf+=1;
		*buf=100;
		buf+=1;
		break;
	}
}

frag::frag(short self)
{
	char obsc[16]; //Object name scratchpad

	this->self=-1;
	if(!(self>=0 && self<ISIZE))
		throw error("Given frag index for loading out of range");
	sprintf(obsc,"Frag%hd",self);
	database::switchobj(obsc);
	load();
	this->self=self;
	frags[self]=this;
}

void frag::physics()
{
	rng--;
	if(rng<1)
	{
		delete this;
		return;
	}
	if(typ==2)
		rot=(rot+3)%36;
	if(trg && (typ==1 || typ==2))
	{
		if(trg->colldetect(loc,mov))
		{
			trg->hit(pow,loc,mov,own);
			delete this;
			return;
		}
	}
	
	loc.x+=mov.xx;
	loc.y+=mov.yy;
}

void frag::home()
{
	vect trv; //Target vector
	pol trp; //Target polar

	trv.xx=trg->loc.x-loc.x;
	trv.yy=trg->loc.y-loc.y;
	trp=calc::vecttopol(trv);
	if(trp.rad!=0)
	{
		trv.xx=(5*trv.xx*trck)/trp.rad;
		trv.yy=(5*trv.yy*trck)/trp.rad;
	}
	if(trv.xx>mov.xx)
	{
		mov.xx+=trck/5;
		if(trv.xx<mov.xx)
			mov.xx=trv.xx;		
		}
	if(trv.xx<mov.xx)
	{
		mov.xx-=trck/5;
		if(trv.xx>mov.xx)
			mov.xx=trv.xx;
	}
	if(trv.yy>mov.yy)
	{
		mov.yy+=trck/5;
		if(trv.yy<mov.yy)
			mov.yy=trv.yy;
	}
	if(trv.yy<mov.yy)
	{
		mov.yy-=trck/5;	
		if(trv.yy>mov.yy)
			mov.yy=trv.yy;		
	}
}

void frag::save()
{
	database::putvalue("Type",typ);
	database::putvalue("Sprite",spr);
	if(trg)
		database::putvalue("Target",trg->self);
	if(own)
		database::putvalue("Owner",own->self);
	database::putvalue("XLoc",loc.x);
	database::putvalue("YLoc",loc.y);
	database::putvalue("XVect",mov.xx);
	database::putvalue("YVect",mov.yy);
	database::putvalue("Rotation",rot);
	database::putvalue("Power",pow);
	database::putvalue("Tracking",trck);
	database::putvalue("Range",rng);
}

void frag::load()
{
	typ=database::getvalue("Type");
	spr=database::getvalue("Sprite");
	trg=ship::get(database::getvalue("Target"));
	own=ship::get(database::getvalue("Owner"));
	loc.x=database::getvalue("XLoc");
	loc.y=database::getvalue("YLoc");
	mov.xx=database::getvalue("XVect");
	mov.yy=database::getvalue("YVect");
	rot=database::getvalue("Rotation");
	pow=database::getvalue("Power");
	trck=database::getvalue("Tracking");
	rng=database::getvalue("Range");
}

frag* frag::frags[ISIZE];
short frag::sel;
