/*
	game.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <stdio.h>
#include "alliance.h"
#include "calc.h"
#include "server.h"
#include "ship.h"
#include "planet.h"
#include "frag.h"
#include "client.h"
#include "ticker.h"
#include "error.h"
#include "interface.h"
#include "camera.h"
#include "presence.h"
#include "graphic.h"
#include "database.h"

class game //This modules allows access to start a general game type
{
	public:
	static void runheadless(); //Run as a headless server
	static void runlocal(); //Run a local single player game
	static void runclient(char* host); //Just run the client, connecting to given hostname

	private:
	static void save(char* path); //Save the entire game state to a file
	static void load(char* path); //Load the entire game state from a file
};

void game::runheadless() //Run as a headless server
{
	ticker lreg(24); //Loop regulator

	lreg.start();
	try
	{
		server::start();
		alliance::maketerritories();
		for(short i=0;i<400;i++)
			planet::shipyards();
		while(true)
		{
			server::cycle();
			ship::simulateall();
			frag::simulateall();
			ship::behaveall();
			planet::shipyards();
			lreg.tick();
		}
	}
	catch(error it)
	{
		server::stop();
		ship::purgeall();
		planet::purgeall();
		frag::purgeall();
		throw it;
	}
}

void game::runlocal()
{
	ticker lreg(24); //Loop regulator
	short fdrp; //Frame drop rate
	short fcnt; //Frame drop counter

	long tcks=0;

	lreg.start();
	fcnt=0;
	fdrp=100;
	try
	{
		server::start();
		try
		{
			load("localuniverse.svd");
		}
		catch(error it)
		{
			alliance::maketerritories();
			for(short i=0;i<400;i++)
				planet::shipyards();
		}
		client::connect("127.0.0.1");
		while(true)
		{
			tcks=(tcks+1)%24;
			interface::poll();
			presence::controls();
			server::cycle();
			presence::interpolateall();
			client::poll();
			camera::update();
			fcnt+=100;
			if(fcnt>=fdrp)
			{
				fcnt-=fdrp;
				camera::render();
				interface::render();
				presence::render();
				graphic::blit();
			}
			ship::simulateall();
			frag::simulateall();
			ship::behaveall();
			planet::shipyards();
			if(lreg.afps<23)
				fdrp++;
			if(lreg.afps>23.9 && fdrp>100)
				fdrp--;
			lreg.tick();
		}
	}
	catch(error it)
	{
		if(presence::me)
			save("localuniverse.svd");
		client::stop();
		server::stop();
		ship::purgeall();
		planet::purgeall();
		frag::purgeall();
		presence::purgeall();
		camera::turnoff();
		camera::unbind();
		interface::printtomesg(it.str);
	}
}

void game::runclient(char* host)
{
	ticker lreg(24); //Loop regulator
	short fdrp; //Frame drop rate
	short fcnt; //Frame drop counter

	lreg.start();
	fcnt=0;
	fdrp=100;
	try
	{
		client::connect(host);
		while(true)
		{
			interface::poll();
			presence::controls();
			presence::interpolateall();
			client::poll();
			camera::update();
			fcnt+=100;
			if(fcnt>=fdrp)
			{
				fcnt-=fdrp;
				camera::render();
				interface::render();
				presence::render();
				graphic::blit();
			}
			if(lreg.afps<23)
				fdrp++;
			if(lreg.afps>23.9 && fdrp>100)
				fdrp--;
			lreg.tick();
		}
	}
	catch(error it)
	{
		client::stop();
		presence::purgeall();
		camera::turnoff();
		camera::unbind();
		interface::printtomesg(it.str);
	}
}

void game::save(char* path)
{
	database::openwriter(path);
	planet::saveall();
	ship::saveall();
	frag::saveall();
	database::closewriter();
}

void game::load(char* path)
{
	database::openreader(path);
	planet::loadall();
	ship::loadall();
	frag::loadall();
	database::closereader();
}
