/*
	graphic.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <SDL.h>
#include "SDL_rotozoom.h"
#include "SDL_gfxPrimitives.h"
#include "calc.h"
#include "error.h"
#include "interface.h"

struct sbox //Bounding box on screen
{
	short x,y,w,h; //Co-ordinates, width and height
};

class graphic //A sprite object
{
	public:
	static const short ISIZE=512; //Graphics index size
	enum {LOGO=0,FIRE=1,NAV=2,WARP=3,PANEL=4,GRID=5,POS=6,TRG=7}; //Special gfx index/file numbers
	enum {BLACK=0,RED=1,LIGHTRED=2,GREEN=3,LIGHTGREEN=4,BLUE=5,LIGHTBLUE=6,YELLOW=7,ORANGE=8,PURPLE=9,GREY=10,DARKGREY=11,WHITE=12}; //Colour values

	static void init(); //Initialise the graphic system data
	static void setup(bool big,bool full); //Setup the screen, with big as true sets up an extra large window, full for fullscreen
	static void blit(); //Commit all graphics to the screen
	static graphic* get(short indx); //Fetch a graphic sprite by index
	static void string(char* str,short x,short y); //Render a string to the screen, string and co-ordinates given
	static void box(sbox* box,short col); //Render a box of given dimensions and color to the screen
	static void clip(sbox* box); //Set the graphics clipping box
	static void pix(short x,short y,short col); //Render a pixel to the screen
	static sbox dimension(); //Return the dimensions of the screen as an sbox
	static void line(short x1,short x2,short y1,short y2,short col); //Draw a line with given colour and co-ordinates

	void draw(short x,short y,short rot,short zout,short haze,bool trg); //Draw this graphic at given co-ordinates, rotation and zoom-out, with given haze percentage. trg determines if targetting crosshairs are drawn

	private:
	graphic(short indx); //Constructor to create a clean graphic object
	void load(); //Load sprite from file
	void calculate(short rot,short zout); //Calculate a rotation and zoom

	static graphic* graphics[ISIZE]; //Index of available graphic objects (sprites)
	static SDL_Surface* screen; //Main drawing surface
	static SDL_Surface* font; //Font store
	static SDL_Surface* cloak; //Cloaking haze image
	static unsigned long cols[16]; //Colour keys, defined at initialisation

	short self; //Self index of this sprite
	SDL_Surface* rots[36][4]; //Rotations and zooms, calculated as needed (rot[0][0] loaded from disk)
	bool imem; //In memory and loaded?
	bool miss; //Sprite missing? Keep this as a record and try not to load it again
};

void graphic::init()
{
	screen=NULL;
	for(short i=0;i<ISIZE;i++)
		graphics[i]=NULL;
}

void graphic::setup(bool big,bool full)
{
	Uint32 flags; //Flags for setting video;
	SDL_Surface* tmp; //Temporary holding place while sprites are converted
	char* path; //Path to load bmp from

	if(SDL_InitSubSystem(SDL_INIT_VIDEO)==-1)
		throw new error(SDL_GetError());
	flags=SDL_DOUBLEBUF;
	if(full)
		flags=flags|SDL_FULLSCREEN;
	if(big)
		screen=SDL_SetVideoMode(800,600,0,flags);
	else
		screen=SDL_SetVideoMode(600,400,0,flags);
	if(!screen)
		throw error(SDL_GetError());
	SDL_WM_SetCaption("Star Voyager","Star Voyager");
	SDL_ShowCursor(0);

	cols[BLACK]=SDL_MapRGB(screen->format,0,0,0);
	cols[RED]=SDL_MapRGB(screen->format,255,0,0);
	cols[LIGHTRED]=SDL_MapRGB(screen->format,255,100,100);
	cols[GREEN]=SDL_MapRGB(screen->format,0,255,0);
	cols[LIGHTGREEN]=SDL_MapRGB(screen->format,100,255,100);
	cols[BLUE]=SDL_MapRGB(screen->format,0,0,255);
	cols[LIGHTBLUE]=SDL_MapRGB(screen->format,100,100,255);
	cols[YELLOW]=SDL_MapRGB(screen->format,255,255,0);
	cols[ORANGE]=SDL_MapRGB(screen->format,255,100,100);
	cols[PURPLE]=SDL_MapRGB(screen->format,255,0,255);
	cols[GREY]=SDL_MapRGB(screen->format,180,180,180);
	cols[DARKGREY]=SDL_MapRGB(screen->format,64,64,64);
	cols[WHITE]=SDL_MapRGB(screen->format,255,255,255);

	path=new char[strlen(DATADIR)+32];
	sprintf(path,"%s/gfx/font.bmp",DATADIR);
	font=SDL_LoadBMP(path);
	delete(path);
	if(!font)
		throw error(SDL_GetError());
	tmp=SDL_ConvertSurface(font,screen->format,SDL_SWSURFACE);
	SDL_FreeSurface(font);
	font=tmp;
	SDL_SetColorKey(font,SDL_SRCCOLORKEY|SDL_RLEACCEL,cols[BLACK]);

	path=new char[strlen(DATADIR)+32];
	sprintf(path,"%s/gfx/haze.bmp",DATADIR);
	cloak=SDL_LoadBMP(path);
	delete(path);
	if(!cloak)
		throw error(SDL_GetError());
	tmp=SDL_ConvertSurface(cloak,screen->format,SDL_SWSURFACE);
	SDL_FreeSurface(cloak);
	cloak=tmp;
	SDL_SetColorKey(cloak,SDL_SRCCOLORKEY|SDL_RLEACCEL,cols[WHITE]);
}

void graphic::blit()
{
	SDL_Flip(screen);
}

graphic* graphic::get(short indx)
{
	if(!(indx>=0 && indx<ISIZE))
		return NULL;
	if(!graphics[indx])
	{
		graphics[indx]=new graphic(indx);
	}
	return graphics[indx];
}

void graphic::string(char* str,short x,short y)
{
	SDL_Rect srct,drct; //Source rect and destination rect
	short i; //Position in text
	char c; //Character in use
	char l; //Letter to print

	i=0;
	srct.y=1;
	srct.w=5;
	srct.h=5;
	drct.x=x;
	drct.y=y;
	drct.w=5;
	drct.h=5;
	while(str[i]!=0)
	{
		c=str[i];
		if(c>32 && c<127)
		{
			l=c-33;
			srct.x=l*6+1;
			SDL_BlitSurface(font,&srct,screen,&drct);
			drct.x+=6;		
		}
		if(c==' ')
			drct.x+=6;
		i++;
	}
	srct.x=x;
	srct.y=y;
	srct.w=drct.x-x;
	srct.h=5;
}

void graphic::box(sbox* box,short col)
{
	SDL_Rect rect; //Rectangle to draw

	rect.x=box->x;
	rect.y=box->y;
	rect.w=box->w;
	rect.h=box->h;
	SDL_FillRect(screen,&rect,cols[col]);
}

void graphic::clip(sbox* box)
{
	SDL_Rect rect; //Rectangle to clip with

	rect.x=box->x;
	rect.y=box->y;
	rect.w=box->w;
	rect.h=box->h;
	SDL_SetClipRect(screen,&rect);
}

void graphic::pix(short x,short y,short col)
{
	fastPixelColor(screen,x,y,cols[col]);
}

sbox graphic::dimension()
{
	sbox out; //Value to output

	out.x=0;
	out.y=0;
	out.w=screen->w;
	out.h=screen->h;
	return out;
}

void graphic::line(short x1,short x2,short y1,short y2,short col)
{
	if(col>=0 && col<16)
	{
		lineColor(screen,(Sint16)x1,(Sint16)x2,(Sint16)y1,(Sint16)y2,cols[col]);
	}
}

void graphic::draw(short x,short y,short rot,short zout,short haze,bool trg)
{
	graphic* tspr; //Targetting sprite
	SDL_Rect dst; //Destination rect
	short hw,hh; //Half-width and half-height, for centering

	if(miss)
		return;
	if(!(rot>=0 && rot<36 && zout>=0 && zout<4))
		return;
	if(!rots[rot][zout])
		calculate(rot,zout);
	hw=(rots[rot][zout]->w)/2;
	hh=(rots[rot][zout]->h)/2;
	dst.x=x-hw;
	dst.y=y-hh;
	SDL_BlitSurface(rots[rot][zout],NULL,screen,&dst);
	if(trg)
	{
		tspr=get(TRG);
		if(tspr)
		{
			tspr->draw(x,y-hh,0,0,0,false);
			tspr->draw(x,y+hh,18,0,0,false);
			tspr->draw(x-hw,y,27,0,0,false);
			tspr->draw(x+hw,y,9,0,0,false);
		}
	}
	if(haze>0)
	{
		for(short i=0;i<haze;i++)
		{
			dst.x=x+calc::rnd(hw*2)-hw;
			dst.y=y+calc::rnd(hh*2)-hh;
			SDL_BlitSurface(cloak,NULL,screen,&dst);
		}
	}
}

graphic::graphic(short indx)
{
	self=indx;
	miss=false;
	for(short i=0;i<36;i++)
		for(short j=0;j<4;j++)
			rots[i][j]=NULL;
	try
	{
		load();
	}
	catch(error it)
	{
		miss=true;
	}
}

void graphic::load()
{
	SDL_Surface* tmp; //Temporary holding place while sprites are converted
	char* path; //Path to load sprite from

	path=new char[strlen(DATADIR)+32];
	sprintf(path,"%s/gfx/%i.bmp",DATADIR,self);
	rots[0][0]=SDL_LoadBMP(path);
	delete(path);
	if(!rots[0][0])
		throw error(SDL_GetError());
	tmp=SDL_ConvertSurface(rots[0][0],screen->format,SDL_SWSURFACE);
	SDL_FreeSurface(rots[0][0]);
	rots[0][0]=tmp;
	SDL_SetColorKey(rots[0][0],SDL_SRCCOLORKEY|SDL_RLEACCEL,cols[BLACK]);
}

void graphic::calculate(short rot,short zout)
{
	short ang; //Angle to rotate by

	ang=360-rot*10;
	rots[rot][zout]=rotozoomSurface(rots[0][0],ang,1.0-((double)zout/7.0),0);
	if(!rots[rot][zout])
		throw error(SDL_GetError());
	SDL_SetColorKey(rots[rot][zout],SDL_SRCCOLORKEY|SDL_RLEACCEL,cols[BLACK]);
}

graphic* graphic::graphics[ISIZE];
SDL_Surface* graphic::screen;
SDL_Surface* graphic::font;
SDL_Surface* graphic::cloak;
unsigned long graphic::cols[16];
