/*
	interface.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <SDL.h>
#include <stdio.h>
#include <stdarg.h>
#include "graphic.h"
#include "error.h"
#include "protocol.h"
#include "calc.h"

class interface //User interface module
{
	public:
	static void init(); //Initialise datastructures
	static void setup(); //Setup layout and stuff
	static void poll(); //Main interface cycle, polling for keyboard and other input
	static void printtocons(char* fmt,...); //Print to the game console
	static void spritetocons(graphic* spr); //Draw an underlay sprite to the console
	static void printtomesg(char* fmt,...); //Print to the message tickertape
	static void render(); //Update the interface to screen
	static bool getline(char* put,bool hide); //Make sure the editing buffer is active and return true if text has been entered and put it into the given string buffer, hide determining password field style

	static sbox viewb,radarb,barsb,panelb; //Bounding boxes for radar, main view, indicator bars and the rightside panel
	static bool inp; //In input mode?
	static short lkey; //Keysym of last key pressed
	static unsigned char lasc; //Ascii value of last key pressed
	static unsigned char* keys; //Pointer to keyboard state

	private:
	static void lineedit(); //Function to handle line-editing

	static sbox consb,mesgb,editb; //Bounding boxes for console, message and editing panes
	static char* cons; //Console text
	static char* mesg[4]; //Message ticker tape lines
	static bool scr; //Message just scrolled?
	static char edit[65]; //Editing buffer
	static bool pwd; //In password mode for editing buffer?
	static bool ent; //Line entering completed?
	static graphic* cspr; //Underlay console sprite
};

void interface::init()
{
	cons=NULL;
	for(short i=0;i<4;i++)
		mesg[i]=NULL;
	scr=false;
	edit[0]='\0';
	inp=false;
	ent=false;
	cspr=NULL;
}

void interface::setup()
{
	sbox tot; //Total screen area

	tot=graphic::dimension();

	viewb.x=tot.x;
	viewb.y=tot.y;
	viewb.w=tot.h;
	viewb.h=tot.h;

	panelb.x=viewb.x+viewb.w;
	panelb.y=tot.y;
	panelb.w=(tot.x+tot.w)-panelb.x;
	panelb.h=400;

	consb.x=panelb.x+5;
	consb.y=panelb.y+200;
	consb.w=(panelb.x+panelb.w)-consb.x-5;
	consb.h=panelb.h-214;

	radarb.x=panelb.x+5;
	radarb.y=panelb.y+25;
	radarb.w=150;
	radarb.h=radarb.w;

	barsb.x=panelb.x+panelb.w-38;
	barsb.y=panelb.y+75;
	barsb.w=32;
	barsb.h=100;

	editb.x=consb.x;
	editb.y=consb.y+consb.h+1;
	editb.w=consb.w;
	editb.h=10;

	mesgb.x=viewb.x+7;
	mesgb.y=viewb.y+viewb.h-35;
	mesgb.w=viewb.w-20;
	mesgb.h=viewb.y+viewb.h-mesgb.y-8;

	cons=new char[(consb.w/7)*(consb.h/7)];
	cons[0]='\0';

	graphic::clip(&tot);
	graphic::box(&tot,graphic::BLACK);

	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,SDL_DEFAULT_REPEAT_INTERVAL);
}

void interface::poll()
{
	SDL_Event evnt; //Event to be polled
	int n; //Unused area for the argument for SDL_GetKeyState

	lkey=-1;
	lasc=0;
	SDL_PumpEvents();
	keys=SDL_GetKeyState(&n);
	while(lkey==-1 && SDL_PollEvent(&evnt))
	{
		if(evnt.type==SDL_KEYDOWN)
		{
			lkey=evnt.key.keysym.sym;
			lasc=evnt.key.keysym.unicode;
			if(inp)
				lineedit();
		}
		if(evnt.type==SDL_QUIT)
			throw error("User requested quit");
	}
}

void interface::printtocons(char* fmt,...)
{
	va_list fmts; //To help parsing the vargs

	inp=false;
	edit[0]='\0';
	cspr=NULL;
	va_start(fmts,fmt);
	vsprintf(cons,fmt,fmts);
	va_end(fmts);
}

void interface::spritetocons(graphic* spr)
{
	inp=false;
	edit[0]='\0';
	cspr=spr;
}

void interface::printtomesg(char* fmt,...)
{
	va_list fmts; //To help parsing the vargs
	char txt[1024]; //Temporary text buffer

	if(mesg[0])
		delete mesg[0];
	mesg[0]=mesg[1];
	mesg[1]=mesg[2];
	mesg[2]=mesg[3];
	mesg[3]=NULL;
	va_start(fmts,fmt);
	vsprintf(txt,fmt,fmts);
	va_end(fmts);
	mesg[3]=new char[strlen(txt)+1];
	strcpy(mesg[3],txt);
	scr=true;
}

void interface::render()
{
	char* p1;
	char* p2; //Pointers for parsing text into console
	short y; //Y co-ordinate for writing to console
	char line[256]; //Line of console for scratchpad operations
	short cwid; //Console width (in characters)
//	sbox eb; //Editing border

	graphic::clip(&mesgb);
	if(scr)
	{
		graphic::box(&mesgb,graphic::BLACK);
		scr=false;
	}
	for(short i=0,y=mesgb.y;i<4;i++,y+=7)
		if(mesg[i])
			graphic::string(mesg[i],mesgb.x,y);

	graphic::clip(&consb);
	graphic::box(&consb,graphic::BLACK);

	if(cspr)
		cspr->draw(consb.x+consb.w/2,consb.y+consb.h/2,0,0,0,false);
	cwid=consb.w/6;
	y=consb.y;
	p1=cons;
	p2=cons;
	while(*p2)
	{
		if(*p2=='\0' || *p2=='\n' || p2-p1==cwid || p2-p1>=255)
		{
			memcpy(line,p1,p2-p1+1);
			line[p2-p1+1]='\0';
			graphic::string(line,consb.x,y);
			y+=7;
			p1=p2+1;
			if(*p2=='\0')
				break;
		}
		p2++;
	}

	graphic::clip(&editb);
	if(inp)
		graphic::box(&editb,graphic::RED);
	else
		graphic::box(&editb,graphic::BLACK);
	if(pwd)
	{
		for(short i=0,j=strlen(edit);i<=j;i++)
		{
			if(i==j)
				line[i]='\0';
			else
				line[i]='*';
		}
		graphic::string(line,editb.x+2,editb.y+2);
	}
	else
		graphic::string(edit,editb.x+2,editb.y+2);
}

sbox interface::consb;

bool interface::getline(char* put,bool hide)
{
	pwd=hide;
	if(!inp)
	{
		inp=true;
		ent=false;
		edit[0]='\0';
	}
	if(ent)
	{
		strcpy(put,edit);
		//edit[0]='\0';
		inp=false;
		ent=false;
		return true;
	}
	else
		return false;
}

sbox interface::viewb,interface::radarb,interface::barsb,interface::panelb;
bool interface::inp;
short interface::lkey;
unsigned char interface::lasc;
unsigned char* interface::keys;

void interface::lineedit() //Function to handle line-editing
{
	if(lkey==SDLK_RETURN)
	{
		ent=true;
		return;
	}
	if(lkey==SDLK_BACKSPACE)
	{
		for(short i=1;i<64;i++)
		{
			if(edit[i]=='\0')
			{
				edit[i-1]='\0';
				break;
			}
		}
	}
	if(lasc>=32 && lasc<127)
	{
		for(short i=0;i<64;i++)
		{
			if(edit[i]=='\0')
			{
				edit[i]=lasc;
				edit[i+1]='\0';
				break;
			}
		}
	}
}

sbox interface::mesgb;
sbox interface::editb;
char* interface::cons;
char* interface::mesg[4];
bool interface::scr;
char interface::edit[65];
bool interface::pwd;
bool interface::ent;
graphic* interface::cspr;
