/*
	planet.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <stdio.h>
#include "calc.h"
#include "error.h"
#include "alliance.h"
#include "database.h"
#include "equip.h"
#include "ship.h"
#include "protocol.h"

class planet //Planet object
{
	public:
	static const short ISIZE=256; //Count of available planets in the index
	enum {STAR=0,INHABITED=1,UNINHABITED=2}; //Planetary object types

	planet(char* nam,cord put,short typ,alliance* all); //Spawn a new planet
	~planet(); //Destructor, resolves dependencies

	static void init(); //Initialise datastructures for the planet module
	static void purgeall(); //Cleans up and purges planet data, used after a game
	static planet* get(short indx); //Return the planet of given index
	static planet* pick(alliance* tali); //Pick a random planet of given alliance
	static planet* pickally(alliance* tali); //Pick a random planet allied to given alliance
	static planet* pickhostile(alliance* tali); //Pick a random planet hostile to given alliance
	static bool masslock(cord loc); //Return whether or not given location is mass-locked by planets
	static void saveall(); //Save all the planets to the database
	static void loadall(); //Load all the planets from the database
	static void generatename(char* put); //Generate a planetary name, put it in put
	static void shipyards(); //Randomly spawn ships

	short serverreq(char* txt,short cmod,short opr,ship* mshp); //Handles a server request for information/action from this planet, with the given comm mode, operand and player's ship, writing the text into txt and returning the sprite index (-1 if n/a)
	void netout(short typ,unsigned char* buf); //Get type of data from planet into a network buffer

	short self; //Self index
	short spr; //Sprite to display for this
	cord loc; //Location
	alliance* all; //Alliance
	short typ; //Planet type 1:Sun 2:Uninhabited 3:Inhabited

	private:
	planet(short self); //Constructor, given self index loads the planet from the database

	void save(); //Save this planet to the database
	void load(); //Load this planet from the database
	void shipyard(); //Spanw a random (but appropriate) ship at this planet

	static planet* planets[ISIZE]; //Planets in the index

	char nam[65]; //Name
	short rot; //Version of sprite to use
	equip* sold[8]; //Sold equipment
};

planet::planet(char* nam,cord put,short typ,alliance* all)
{
	short slo=100,shi=103,ulo=110,uhi=116,ilo=120,ihi=124; //Boundings for planet sprites

	self=-1;
	for(short i=0;i<ISIZE && self==-1;i++)
	{
		if(!planets[i])
		{
			self=i;
			planets[i]=this;
		}
	}
	if(self==-1)
		throw error("No free slots in the planet index");
	sprintf(this->nam,"%s",nam);
	this->loc=put;
	this->all=all;
	rot=calc::rnd(36);
	if(typ==STAR)
		spr=slo+calc::rnd(shi-slo+1);
	if(typ==UNINHABITED)
		spr=ulo+calc::rnd(uhi-ulo+1);
	if(typ==INHABITED)
		spr=ilo+calc::rnd(ihi-ilo+1);
	this->typ=typ;
	for(short j=0;j<8;j++)
		sold[j]=all->getequip();
}

planet::~planet()
{
	if(self!=-1)
		planets[self]=NULL;
}

void planet::init()
{
	for(short i=0;i<ISIZE;i++)
		planets[i]=NULL;
}

void planet::purgeall()
{
	for(short i=0;i<ISIZE;i++)
		if(planets[i])
			delete planets[i];
}

planet* planet::get(short indx)
{
	if(indx>=0 && indx<ISIZE)
	{
		if(planets[indx])
			return planets[indx];
		else
			return NULL;
	}
	else
	{
		return NULL;
	}
}

planet* planet::pick(alliance* tali)
{
	for(short i=0,j=0;i<ISIZE;i++)
	{
		j=calc::rnd(ISIZE);
		if(planets[j] && planets[j]->typ!=STAR && planets[j]->all==tali)
			return planets[j];
	}
	return NULL;
}

planet* planet::pickally(alliance* tali)
{
	for(short i=0,j=0;i<ISIZE;i++)
	{
		j=calc::rnd(ISIZE);
		if(planets[j] && planets[j]->typ!=STAR && !(tali->opposes(planets[j]->all)))
			return planets[j];
	}
	return NULL;
}

planet* planet::pickhostile(alliance* tali)
{
	for(short i=0,j=0;i<ISIZE;i++)
	{
		j=calc::rnd(ISIZE);
		if(planets[j] && planets[j]->typ!=STAR && tali->opposes(planets[j]->all))
			return planets[j];
	}
	return NULL;
}

bool planet::masslock(cord loc)
{
	long dx,dy; //Co-ordinate differences
	
	for(short i=0;i<ISIZE;i++)
	{
		if(planets[i])
		{
			dx=loc.x-planets[i]->loc.x;
			dy=loc.y-planets[i]->loc.y;
			if(dx<10000 && dx>-10000 && dy>-10000 && dy<10000)
				return true;
		}
	}
	return false;
}

void planet::saveall()
{
	char obsc[33]; //Object name scratchpad

	for(short i=0;i<ISIZE;i++)
	{
		if(planets[i])
		{
			sprintf(obsc,"Planet%hd",i);
			database::putobject(obsc);
			planets[i]->save();
		}
	}
}

void planet::loadall()
{
	for(short i=0;i<ISIZE;i++)
	{
		try
		{
			new planet(i);
		}
		catch(error it)
		{
		}
	}
}

void planet::generatename(char* put)
{
	const char n1[][16]={"Vul","Et","Fer","Man","Clo","Jen","Reb","Joy","Dan","Sat","Mat","Cor","Wat","An","Oct","Ack"};
	const char n2[][16]={"tan","can","an","en","in","lat","pil","op","ep","ast","tyn","syn","blat","av","",""};
	const char n3[][16]={"is","ia","ol","ion","on","ath","ur","e","ise","at","","","","","",""};
	short s1,s2,s3; //Syllable selectors

	s1=calc::rnd(16);
	s2=calc::rnd(16);
	s3=calc::rnd(16);
	sprintf(put,"%s%s%s",n1[s1],n2[s2],n3[s3]);
}

void planet::shipyards()
{
	planet* tpln; //Planet to spawn at

	tpln=planets[calc::rnd(ISIZE)];
	if(tpln && tpln->typ==INHABITED)
		tpln->shipyard();

}

short planet::serverreq(char* txt,short cmod,short opr,ship* mshp)
{
	long cost; //Cost of purchase

	if(!mshp)
		return -1;
	switch(cmod)
	{
		case CMOD_SCAN:
		if(opr==-1)
		{
			txt+=sprintf(txt,"%s\n",nam);
			if(mshp->all->opposes(all))
				txt+=sprintf(txt,"Alignment:%s [hostile]\n",all->nam);
			else
				txt+=sprintf(txt,"Alignment:%s\n",all->nam);
			switch(typ)
			{
				case STAR:
				txt+=sprintf(txt,"Star\n");
				break;

				case UNINHABITED: 
				txt+=sprintf(txt,"Uninhabited planet\n");
				break;

				case INHABITED:
				txt+=sprintf(txt,"Inhabited planet\n");
				break;
			}
			txt+=sprintf(txt,"\n[1] Lay in a course\n");
			return spr;
		}
		if(opr==1 && this==mshp->plnt)
		{
			mshp->stat=ship::STAT_PLANET;
		}
		break;

		case CMOD_HAIL:
		if(mshp->see(this))
		{
			if(
			typ!=INHABITED ||		
			all->trad==alliance::TRADE_NONE ||
			(all->trad==alliance::TRADE_CLOSED && mshp->all!=this->all) ||
			(all->trad==alliance::TRADE_FRIENDS && this->all->opposes(mshp->all))
			)
			{
				txt+=sprintf(txt,"Hailing %s\n\nNo response",nam);
				return -1;
			}
		}
		else
		{
			txt+=sprintf(txt,"Out of range");
			return -1;
		}
		if(opr==-1)
		{
			txt+=sprintf(txt,"Hailing %s\n\nServices\n\n",nam);
			cost=mshp->purchase(ship::PRCH_FUEL,all->ripo,false);
			if(cost)
				txt+=sprintf(txt,"[1] Refuel\nCost: %ld C\n",cost);
			cost=mshp->purchase(ship::PRCH_HULL,all->ripo,false);
			if(cost)
				txt+=sprintf(txt,"[2] Repair hull\nCost: %ld C\n",cost);
			cost=mshp->purchase(ship::PRCH_ARMS,all->ripo,false);
			if(cost)
				txt+=sprintf(txt,"[3] Rearm one magazine\nCost: %ld C\n",cost);
			txt+=sprintf(txt,"[4] Purchase equipment\n");
		}
		if(opr==1 || opr==2 || opr==3)
		{
			if(opr==1)
				cost=mshp->purchase(ship::PRCH_FUEL,all->ripo,false);
			if(opr==2)
				cost=mshp->purchase(ship::PRCH_HULL,all->ripo,false);
			if(opr==3)
				cost=mshp->purchase(ship::PRCH_ARMS,all->ripo,false);
			if(cost>0)
			{
				if(mshp->transport(this,true))
				{
					if(opr==1)
						mshp->purchase(ship::PRCH_FUEL,all->ripo,true);
					if(opr==2)
						mshp->purchase(ship::PRCH_HULL,all->ripo,true);
					if(opr==3)
						mshp->purchase(ship::PRCH_ARMS,all->ripo,true);
				}
				else
					txt+=sprintf(txt,"Cannot establish transporter lock");
			}
			else
				txt+=sprintf(txt,"Cannot purchase that item");
		}
		break;

		case CMOD_REFIT:
		if(mshp->see(this))
		{
			if(typ!=INHABITED)
			{
				txt+=sprintf(txt,"Hailing %s\n\n",nam);
				txt+=sprintf(txt,"No response");
				return -1;
			}
		}
		else
		{
			txt+=sprintf(txt,"Hailing %s\n\n",nam);
			txt+=sprintf(txt,"Out of range");
			return -1;
		}
		if(opr==-1)
		{
			txt+=sprintf(txt,"Hailing %s\n\nEquipment\n(affordable and installable)\n\n",nam);
			for(short i=0;i<8;i++)
			{
				if(sold[i])
				{
					cost=mshp->purchase(sold[i],all->ripo,false);
					if(cost>0)
						txt+=sprintf(txt,"[%hd] %s\nCost: %ld C  Mass: %hd\n",i+1,sold[i]->nam,cost,sold[i]->mss);
				}
			}
			txt+=sprintf(txt,"\nAvailable mass: %hd\n",mshp->freemass());
		}
		if(opr>=1 && opr<=8)
		{
			cost=mshp->purchase(sold[opr-1],all->ripo,false);
			if(cost>0)
			{
				if(mshp->transport(this,true))
					mshp->purchase(sold[opr-1],all->ripo,true);
				else
					txt+=sprintf(txt,"Cannot establish transporter lock");
					
			}
			else
				txt+=sprintf(txt,"Cannot purchase that item");
		}
		break;

		default:
		break;
	}
	return -1;
}

void planet::netout(short typ,unsigned char* buf)
{
	buf[0]=typ;
	buf+=1;

	calc::shorttodat(planet2pres(self),buf);
	buf+=2;
	switch(typ)
	{
		case SERV_NEW:
		*buf=PT_PLANET;
		buf+=1;
		calc::shorttodat(spr,buf);
		buf+=2;
		calc::shorttodat(-1,buf);
		buf+=2;
		break;

		case SERV_NAME:
		sprintf((char*)buf,"%s",nam);
		buf+=64;
		sprintf((char*)buf,"%s",all->nam);
		buf+=64;
		break;

		case SERV_UPD:
		calc::longtodat(loc.x,buf);
		buf+=4;
		calc::longtodat(loc.y,buf);
		buf+=4;
		calc::longtodat(0,buf);
		buf+=4;
		calc::longtodat(0,buf);
		buf+=4;
		calc::shorttodat(rot*10,buf);
		buf+=2;
		*buf=0;
		buf+=1;
		*buf=100;
		buf+=1;
		break;
	}
}

planet::planet(short self)
{
	char obsc[16]; //Object name scratchpad

	this->self=-1;
	if(planets[self])
		throw error("Given planet index slot already in use");
	sprintf(obsc,"Planet%hd",self);
	database::switchobj(obsc);
	load();
	this->self=self;
	planets[self]=this;
}

void planet::save()
{
	char atsc[33]; //Attribute scratchpad

	database::putvalue("Name",nam);
	if(spr)
		database::putvalue("Sprite",spr);
	database::putvalue("SpriteRot",rot);
	if(spr)
		database::putvalue("Team",all->self);
	database::putvalue("XLoc",loc.x);
	database::putvalue("YLoc",loc.y);
	database::putvalue("Type",typ);
	for(short i=0;i<8;i++)
	{
		sprintf(atsc,"Sold%hd",i);
		if(sold[i])
			database::putvalue(atsc,sold[i]->self);
		else
			database::putvalue(atsc,-1);
	}
}

void planet::load()
{
	char atsc[33]; //Attribute scratchpad

	database::getvalue("Name",nam);
	spr=database::getvalue("Sprite");
	rot=database::getvalue("SpriteRot");
	all=alliance::get(database::getvalue("Team"));
	loc.x=database::getvalue("XLoc");
	loc.y=database::getvalue("YLoc");
	typ=database::getvalue("Type");
	for(short i=0;i<8;i++)
	{
		sprintf(atsc,"Sold%hd",i);
		sold[i]=equip::get(database::getvalue(atsc));
	}
}

void planet::shipyard()
{
	short spw; //Ship type to spawn
	cord put; //Location to spawn

	put.x=loc.x+calc::rnd(150)-calc::rnd(150);
	put.y=loc.y+calc::rnd(150)-calc::rnd(150);
	spw=all->sinv[calc::rnd(16)];
	if(spw>=0 && spw<ship::LIBSIZE)
	{
		try
		{
			new ship(put,spw,all);
		}
		catch(error it)
		{
		}
	}
}

planet* planet::planets[ISIZE];
