/*
	presence.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <SDL.h>
#include "calc.h"
#include "graphic.h"
#include "protocol.h"
#include "camera.h"
#include "interface.h"
#include "client.h"
#include "error.h"

class presence //Client side view of the universe and its simple objects
{
	public:
	static const short ISIZE=2048; //Size of presence object index
	static void init(); //Initialise the presence system datastructures
	static void purgeall(); //Purge all presence data after use
	static void feed(unsigned char* buf); //Update/create a presence from a buffer
	static void interpolateall(); //Run interpolation on all the presence objects
	static presence* get(short indx); //Return the presence object of given index
	static void render(); //Render client/presence info, such as realtime information indicators
	static void controls(); //Run the in game ship controls

	void drawat(short sx,short sy); //Draw at the given screen location

	static presence* me; //Presence played by this client
	static presence* trg; //Personal target
	static presence* hl; //Hilight target
	static short hul,pow,shd,ful; //Hull, power, shield and fuel percentages
	static pol vel; //Velocity of self
	static long srng,lrng; //Sensor range and long range

	short self; //Presence object self index
	short typ; //Object type
	char nam[65]; //Name of object
	char anno[65]; //Annotative data, e.g. alliance
	cord loc; //Location
	bool enem; //Is enemy?

	private:
	presence(short self,unsigned char* buf); //Create a new object with a given network buffer input and given self index
	~presence(); //Delete this presence object

	static presence* gettarget(short typ,short dir,box cov,bool out,bool enem); //Flip through targets of given type, flicking either forwards or backwards(dir -1 or +1) through them, within the given bounding box, out refers to outside the box or not, enem if you want hostiles only
	static void updself(unsigned char* buf); //Update self data with given buffer

	void name(unsigned char* buf); //Name the object with given network buffer
	void update(unsigned char* buf);  //Update an object with given buffer
	void interpolate(); //Run motion interpolation on this presence

	static presence* objs[ISIZE]; //Index of presence objects
	static short hltm; //Hilight time remaining

	vect mov; //Motion vector
	graphic* spr; //Sprite
	short col; //Colour (makes it a beam)
	short vis; //Visibility percentage (for cloaking stuff)
	short ang; //Angle of orientation
	short age; //Age of presence object
	presence* link; //Link to another presence, for example owner of a frag
};

void presence::init()
{
	me=NULL;
	trg=NULL;
	hl=NULL;
	hltm=0;
	hul=0;
	pow=0;
	shd=0;
	ful=0;
	srng=0;
	lrng=0;
	vel.rad=0;
	vel.ang=0;
	for(short i=0;i<ISIZE;i++)
		objs[i]=NULL;
}

void presence::purgeall()
{
	for(short i=0;i<ISIZE;i++)
		if(objs[i])
			delete objs[i];
	init();
}

void presence::feed(unsigned char* buf)
{
	short id; //Id to use

	if(buf[0]==SERV_SELF)
	{
		updself(buf);
	}
	if(buf[0]==SERV_NEW)
	{
		id=calc::dattoshort(buf+1);
		if(id>=0 && id<ISIZE)
		{
			if(objs[id])
				delete objs[id];
			objs[id]=new presence(id,buf);
		}
	}
	if(buf[0]==SERV_NAME)
	{
		id=calc::dattoshort(buf+1);
		if(id>=0 && id<ISIZE)
		{
			if(objs[id])
				objs[id]->name(buf);
		}
	}
	if(buf[0]==SERV_UPD)
	{
		id=calc::dattoshort(buf+1);
		if(id>=0 && id<ISIZE)
		{
			if(objs[id])
				objs[id]->update(buf);
		}
	}
	if(buf[0]==SERV_DEL)
	{
		id=calc::dattoshort(buf+1);
		if(id>=0 && id<ISIZE)
			if(objs[id])
				delete objs[id];
	}
	if(buf[0]==SERV_HILIGHT)
	{
		id=calc::dattoshort(buf+1);
		if(id>=0 && id<ISIZE && objs[id])
		{
			hl=objs[id];
			hltm=120;
		}
	}
}

void presence::interpolateall()
{
	if(hltm>0)
	{
		hltm--;
		if(hltm==0)
			hl=NULL;
	}
	for(short i=0;i<ISIZE;i++)
		if(objs[i])
			objs[i]->interpolate();
}

presence* presence::get(short indx)
{
	if(indx>=0 && indx<ISIZE)
	{
		if(objs[indx])
			return objs[indx];
		else
			return NULL;
	}
	else
	{
		return NULL;
	}
}


void presence::render()
{
	sbox bbox; //Box for rendering indicator bar

	graphic::clip(&interface::barsb);
	graphic::box(&interface::barsb,graphic::BLACK);
	
	if(me)
	{
		bbox.w=8;

		bbox.x=interface::barsb.x;
		bbox.y=interface::barsb.y+(interface::barsb.h-(shd*interface::barsb.h)/100);
		bbox.h=interface::barsb.h+interface::barsb.y-bbox.y;
		graphic::box(&bbox,graphic::RED);

		bbox.x+=8;
		bbox.y=interface::barsb.y+(interface::barsb.h-(pow*interface::barsb.h)/100);
		bbox.h=interface::barsb.h+interface::barsb.y-bbox.y;
		graphic::box(&bbox,graphic::GREEN);

		bbox.x+=8;
		bbox.y=interface::barsb.y+(interface::barsb.h-(hul*interface::barsb.h)/100);
		bbox.h=interface::barsb.h+interface::barsb.y-bbox.y;
		graphic::box(&bbox,graphic::WHITE);

		bbox.x+=8;
		bbox.y=interface::barsb.y+(interface::barsb.h-(ful*interface::barsb.h)/100);
		bbox.h=interface::barsb.h+interface::barsb.y-bbox.y;
		graphic::box(&bbox,graphic::BLUE);
	}
}

void presence::controls()
{
	presence* trg; //Target
	
	trg=NULL;
	if(interface::keys[SDLK_LEFT])
		client::action(CLIENT_TURN,-1);
	if(interface::keys[SDLK_RIGHT])
		client::action(CLIENT_TURN,+1);
	if(interface::keys[SDLK_DOWN])
		if(interface::keys[SDLK_LSHIFT] || interface::keys[SDLK_RSHIFT])
			client::action(CLIENT_ACCEL,-2);
		else
			client::action(CLIENT_ACCEL,-1);
	if(interface::keys[SDLK_UP])
		if(interface::keys[SDLK_LSHIFT] || interface::keys[SDLK_RSHIFT])
			client::action(CLIENT_ACCEL,+2);
		else
			client::action(CLIENT_ACCEL,+1);
	if(!interface::inp)
	{
		if(interface::keys[SDLK_SPACE])
			client::action(CLIENT_SHOOT,0);
		if(interface::lkey==SDLK_z)
			client::action(CLIENT_SHOOT,1);
		if(interface::lasc=='0')
			client::action(CLIENT_CONS,0);
		if(interface::lasc>='1' && interface::lasc<='9')
			client::action(CLIENT_CONS,interface::lasc-'1'+1);
		if(interface::lasc=='t')
			trg=gettarget(PT_SHIP,+1,camera::cov,false,false);
		if(interface::lasc=='T')
			trg=gettarget(PT_SHIP,-1,camera::cov,false,false);
		if(interface::lasc=='e')
			trg=gettarget(PT_SHIP,+1,camera::cov,false,true);
		if(interface::lasc=='E')
			trg=gettarget(PT_SHIP,-1,camera::cov,false,true);
		if(interface::lasc=='p')
			trg=gettarget(PT_PLANET,+1,camera::cov,false,false);
		if(interface::lasc=='P')
			trg=gettarget(PT_PLANET,-1,camera::cov,false,false);
		if(trg)
			client::action(CLIENT_TRG,trg->self);
		if(interface::lasc=='q')
			throw error("User requested quit");
		if(interface::keys[SDLK_MINUS])
			camera::radarzoom(-1);
		if(interface::keys[SDLK_EQUALS])
			camera::radarzoom(+1);
	}
	if(interface::lkey>=SDLK_F1 && interface::lkey<=SDLK_F10)
	{
		client::action(CLIENT_CMOD,interface::lkey-SDLK_F1);
	}
}

void presence::drawat(short sx,short sy)
{
	short rot; //Rotation frame to use
	long lx,ly; //Co-ordinates of 'link' presence
	graphic* pspr; //Graphic for marking position while cloaked
	short zm; //Zoom amount to use

	zm=0;
	if(spr)
	{
		rot=((ang+5)/10)%36;
		if(rot<0)
			rot+=36;
		if(vis)
		{
			if(typ==PT_PLANET)
				zm=rot%4;
			if(this==trg)
				spr->draw(sx,sy,rot,zm,100-vis,true);
			else
				spr->draw(sx,sy,rot,zm,100-vis,false);
		}
		else
		{
			pspr=graphic::get(graphic::POS);
			pspr->draw(sx,sy,rot,0,0,false);
		}	
	}
	if(col>=0 && link)
	{
		lx=sx-(mov.xx-link->mov.xx)*((2*age)/2);
		ly=sy-(mov.yy-link->mov.yy)*((2*age)/2);
		graphic::line(sx,sy,lx,ly,col);
	}
}

presence* presence::me;
presence* presence::trg;
presence* presence::hl;
short presence::hul;
short presence::pow;
short presence::shd;
short presence::ful;
pol presence::vel;
long presence::srng,presence::lrng;

presence::presence(short self,unsigned char* buf)
{
	this->self=self;
	age=0;

	buf+=3;

	typ=*buf;
	buf+=1;

	col=calc::dattoshort(buf);
	if(col>=0)
	{
		spr=graphic::get(col);
		col=-1;
	}
	else
	{
		spr=NULL;
		col=-col;
	}
	buf+=2;

	link=presence::get(calc::dattoshort(buf));
	buf+=2;
}

presence::~presence()
{
	if(self>=0 && self<ISIZE)
		objs[self]=NULL;
	if(me==this)
	{
		me=NULL;
		camera::unbind();
	}
	if(trg==this)
		trg=NULL;
	if(hl==this)
		hl=NULL;
}

presence* presence::gettarget(short typ,short dir,box cov,bool out,bool enem)
{
	short st; //Starting index

	if(!me)
		return NULL;
	if(dir>0)
		dir=+1;
	else
		dir=-1;
	if(trg)
		st=trg->self+dir;
	else
		st=0;
	if(st<0)
		st=ISIZE;
	if(st>=ISIZE)
		st=0;
	for(short i=0;i<ISIZE;i++)
	{
		if(objs[st] && objs[st]!=me && objs[st]->typ==typ)
			if(
			objs[st]->loc.x<cov.x2 &&
			objs[st]->loc.x>cov.x1 &&		
			objs[st]->loc.y<cov.y2 &&		
			objs[st]->loc.y>cov.y1
			)
			{
				if(!out && ((objs[st]->enem && enem) || !enem))
					return objs[st];
			}
			else
			{
				if(out && ((objs[st]->enem && enem) || !enem))
					return objs[st];
			}
				
		st=(st+dir)%ISIZE;
		if(st<0)
			st+=ISIZE;
	}
	return NULL;
}

void presence::updself(unsigned char* buf)
{
	short i; //Temp, for resolving other presences

	buf+=1;

	i=calc::dattoshort(buf);
	me=get(i);
	if(me)
	{
		camera::turnon();
		camera::bind(me);
	}
	else
	{
		camera::unbind();
	}
	buf+=2;

	hul=calc::dattoshort(buf);
	buf+=2;

	pow=calc::dattoshort(buf);
	buf+=2;

	shd=calc::dattoshort(buf);
	buf+=2;

	ful=calc::dattoshort(buf);
	buf+=2;

	srng=calc::dattolong(buf);
	buf+=4;

	lrng=calc::dattolong(buf);
	buf+=4;

	i=calc::dattoshort(buf);
	if(i>=0 && i<ISIZE && objs[i])
		trg=objs[i];
	else
		trg=NULL;
	buf+=2;
}

void presence::name(unsigned char* buf)
{
	buf+=3;

	memcpy(nam,buf,64);
	nam[64]='\0';
	buf+=64;

	memcpy(anno,buf,64);
	nam[64]='\0';
	buf+=64;
}

void presence::update(unsigned char* buf)
{
	buf+=3;

	if(age>5)
	{
		loc.x+=calc::dattolong(buf);
		loc.x/=2;
		buf+=4;
		loc.y+=calc::dattolong(buf);
		loc.y/=2;
		buf+=4;
	}
	else
	{
		loc.x=calc::dattolong(buf);
		buf+=4;
		loc.y=calc::dattolong(buf);
		buf+=4;
	}

	mov.xx=calc::dattolong(buf);
	buf+=4;

	mov.yy=calc::dattolong(buf);
	buf+=4;

	ang=calc::dattoshort(buf);
	buf+=2;

	enem=*buf;
	buf+=1;

	vis=*buf;
	if(vis>100)
		vis=100;
	buf+=1;
}

void presence::interpolate()
{
	if(this==me)
	{
		vel=calc::vecttopol(mov);
	}
	loc.x+=mov.xx;
	loc.y+=mov.yy;
	age++;
}

presence* presence::objs[ISIZE];
short presence::hltm;
