/*
	server.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include "sockhelper.h"
#include "calc.h"
#include "error.h"
#include "protocol.h"
#include "ship.h"
#include "frag.h"
#include "planet.h"
#include "alliance.h"
#include "constants.h"

class server //Controls connections to players
{
	public:
	static short const CLIENTS=32; //Maximum number of clients

	static void init(); //Initialise server datastructures
	static void start(); //Start the server listening
	static void stop(); //Stop the server and clean up all data
	static void cycle(); //Run one tick of the server
	static void notifydeath(ship* tshp); //Tells the server system that a given ship has died
	static void hail(ship* fr,ship* to,char* msg); //Send a hail from one ship to another with given message
	static void registernoise(ship* fr,short snd); //Register a noise emitted from the given ship, of given sound index
	static void registersound(ship* to,short snd); //Register a sound of given index, internal to the given ship
	static void registershake(ship* to,short mag); //Command given ship to shake
	static void transfer(ship* fr,ship* to); //Transfer player from one ship to another
	static char* playername(ship* tshp); //Return player name associated with this ship, null otherwise

	private:
	server(short self,TCPsocket sock); //Constructor for a client handler, give it the player number and the new incoming socket
	~server(); //Destructor, closes socket

	static void log(char* fmt,...); //Output to the server log target

	void poll(); //Poll for client data and act on it	
	void action(short typ,short opr); //Process an action from the client, given action type and operand
	void changecmod(short opr); //Change to given client<->server console mode
	void cons(short opr); //Receive a console impulse from the client, number given as operand
	void requestline(bool hide); //Request a line of input from the client console, if secret (password mode) then specify
	void input(); //Register that readline input has completed from the client, it's time to process it
	void printtocons(char* fmt,...); //Print to the client's console display
	void spritetocons(short indx); //Bring up given index of sprite on clients console
	void printtomesg(char* fmt,...); //Print an instant message to the client
	void assign(ship* tshp); //Assign this client to given ship
	void uploads(); //Handle presence uploads to the client
	void uploadplanets(); //Handle planet uploads
	void uploadships(); //Handle ship uploads
	void uploadfrags(); //Handle frag uploads
	void kill(); //Kill this player, and restart his/her game
	void hilight(ship* tshp); //Hilight the given ship to this player

	static server* players[CLIENTS]; //Server objects for each possible client
	static long tcks; //Server ticks so far
	static TCPsocket lstn; //Listener socket
	static bool qsig; //Quit signal?

	short self; //Player number
	TCPsocket sock; //Socket associated with this player connection
	sockhelper* hlpr; //Helper for above
	char nam[33]; //Player name
	bool auth; //Authorised yet?
	ship* mshp; //Ship associated with player
	short cmod; //Current communications mode; just statekeeping
	char inpb[65]; //Incoming text input buffer
	bool acth[32]; //Action history, prevents multiple repetition cheating
	bool shpu[ship::ISIZE];
	bool plnu[planet::ISIZE];
	bool frgu[frag::ISIZE]; //Tags to indicate if object uploading to client has been dealt with
	short lsnd; //Last sound registered, to prevent repetitions
	short foc; //Focus target of this player's clientside presence
	short urat; //Upload rate divider
	long cbnd; //Client bandwidth report
	short tout; //Timeout counter
};

struct octets //Structure for extraction of octets from an IP
{
	unsigned char o1;
	unsigned char o2;
	unsigned char o3;
	unsigned char o4;
};

void server::init()
{
	for(short i=0;i<CLIENTS;i++)
		players[i]=NULL;
	lstn=NULL;
}

void server::start()
{
	IPaddress serv; //Server listening interface

	tcks=0;
	SDLNet_ResolveHost(&serv,NULL,PORT);
	lstn=SDLNet_TCP_Open(&serv);
	if(!lstn)
		throw error("Can't open listening socket");
	log("Server started listening on port %hd",PORT);
}

void server::stop()
{
	for(short i=0;i<CLIENTS;i++)
		if(players[i])
			delete players[i];
	if(lstn!=NULL)
	{
		SDLNet_TCP_Close(lstn);
		lstn=NULL;
		log("Server shutdown");
	}
}

void server::cycle()
{
	TCPsocket sock; //Sample socket

	tcks++;
	if(tcks>32766)
		tcks=1;
	sock=NULL;
	sock=SDLNet_TCP_Accept(lstn);
	if(sock)
	{
		for(short i=0;i<CLIENTS;i++)
		{
			if(!players[i])
			{
				players[i]=new server(i,sock);
				break;
			}
		}
	}
	for(short i=0;i<CLIENTS;i++)
	{
		if(players[i])
		{
			try
			{
				players[i]->poll();
				players[i]->uploads();
			}
			catch(error it)
			{
				log("Player %hd dropped from the game: %s",i,it.str);
				delete players[i];
			}
		}
	}
}

void server::notifydeath(ship* tshp)
{
	for(short i=0;i<CLIENTS;i++)
		if(players[i] && players[i]->mshp==tshp)
			players[i]->kill();
}

void server::hail(ship* fr,ship* to,char* msg)
{
	char txt[256]; //Communications text
	char* frnm; //Name of from

	if(!to)
		return;
	frnm=NULL;
	if(fr)
	{
		frnm=playername(fr);
		if(!frnm)
			frnm=fr->cls;
	}

	for(short i=0;i<CLIENTS;i++)
	{	
		if(players[i] && players[i]->mshp==to)
		{
			if(fr)
			{
				sprintf(txt,"%s: %s",frnm,msg);
				players[i]->hilight(fr);
			}
			else
				sprintf(txt,"%s",msg);
			players[i]->printtomesg("%s",txt);
			server::registersound(players[i]->mshp,SND_COMM);
		}
	}

	for(short i=0;i<CLIENTS;i++)
	{
		if(players[i] && players[i]->mshp==fr)
		{
			for(short j=0;j<CLIENTS;j++)
			{
				if(players[j] && players[j]->mshp==to)
				{
					log("Player %hd hailed player %hd: %s",i,j,msg);	
					break;
				}
			}
			break;
		}
	}
}

void server::registernoise(ship* fr,short snd)
{
	unsigned char buf[SERV_NOISE_SZ]; //Buffer for sending sound

	for(short i=0;i<CLIENTS;i++)
	{
		if(players[i] && players[i]->mshp && players[i]->lsnd!=snd && players[i]->mshp->see(fr))
		{	
			players[i]->lsnd=snd;	
			buf[0]=SERV_NOISE;
			calc::shorttodat(snd,buf+1);
			calc::shorttodat(ship2pres(fr->self),buf+3);
			players[i]->hlpr->send(buf,SERV_NOISE_SZ);
		}
	}
}

void server::registersound(ship* to,short snd)
{
	unsigned char buf[SERV_SND_SZ]; //Buffer for sending sound

	for(short i=0;i<CLIENTS;i++)
	{
		if(players[i] && players[i]->mshp==to && players[i]->lsnd!=snd)
		{
			players[i]->lsnd=snd;
			buf[0]=SERV_SND;
			calc::shorttodat(snd,buf+1);
			players[i]->hlpr->send(buf,SERV_SND_SZ);
		}

	}
}

void server::registershake(ship* to,short mag)
{
	unsigned char buf[SERV_SHAKE_SZ]; //Buffer for sending shake

	for(short i=0;i<CLIENTS;i++)
	{
		if(players[i] && players[i]->mshp==to)
		{
			buf[0]=SERV_SHAKE;
			calc::shorttodat(mag,buf+1);
			players[i]->hlpr->send(buf,SERV_SHAKE_SZ);
		}

	}
}

void server::transfer(ship* fr,ship* to)
{
	if(fr && to && fr->ply && !to->ply)
	{
		for(short i=0;i<CLIENTS;i++)
		{
			if(players[i] && players[i]->mshp==fr)
			{
				players[i]->mshp=to;
				fr->enem=NULL;
				fr->plnt=NULL;
				fr->frnd=to;
				fr->stat=ship::STAT_FOLLOW;
				fr->ply=false;
				to->enem=NULL;
				to->plnt=NULL;
				to->frnd=NULL;
				to->stat=ship::STAT_HOLD;
				to->ply=true;
			}
		}
	}
}

char* server::playername(ship* tshp)
{
	for(short i=0;i<CLIENTS;i++)
	{
		if(players[i] && players[i]->mshp==tshp)
			return players[i]->nam;
	}
	return NULL;
}

server::server(short self,TCPsocket sock)
{
	unsigned long ip; //ip of connecting client
	octets* oip; //Ip as octets, for writing to Ip record

	this->self=self;
	this->sock=sock;
	hlpr=NULL;
	hlpr=new sockhelper(sock);
	hlpr->pump();
	auth=false;
	for(short i=0;i<32;i++)
		acth[i]=false;
	mshp=NULL;
	lsnd=-1;
	inpb[0]='\0';
	nam[0]='\0';
	foc=-1;
	for(short i=0;i<ship::ISIZE;i++)
		shpu[i]=false;
	for(short i=0;i<planet::ISIZE;i++)
		plnu[i]=false;
	for(short i=0;i<frag::ISIZE;i++)
		frgu[i]=false;
	ip=SDLNet_TCP_GetPeerAddress(sock)->host;
	oip=(octets*)&ip;
	log("Connection from %hd.%hd.%hd.%hd assigned to player %hd",(short)oip->o1,(short)oip->o2,(short)oip->o3,(short)oip->o4,self);
	urat=10;
	if(oip->o1==127 && oip->o2==0 && oip->o3==0 && oip->o4==1)
		cbnd=999999999;
	else
		cbnd=0;
	tout=0;
}

server::~server()
{
	players[self]=NULL;
	if(hlpr)
		delete hlpr;
	if(mshp)
		delete mshp;
	SDLNet_TCP_Close(sock);
}

void server::log(char* fmt,...)
{
	va_list fmts;

	va_start(fmts,fmt);
	vprintf(fmt,fmts);
	va_end(fmts);
	printf("\n");
}

void server::poll()
{
	ship* ashp; //Ship to potentially assign to
	short typ,opr; //Action description
	unsigned char* buf; //Incoming data buffer
	long sbnd; //Server bandwidth used by this player
	unsigned char fldb[SERV_FLOOD_SZ]; //Flooding buffer

	hlpr->pump();
	if(auth)
	{
		lsnd=-1;
		for(short i=0;i<3;i++)
		{
			buf=hlpr->request(3);
			if(buf)
			{
				typ=((short)buf[0]);
				opr=calc::dattoshort(buf+1);

				action(typ,opr);
				hlpr->suck();
			}
		}
	}
	else
	{
		buf=hlpr->request(6);
		if(buf)
		{
			if(calc::dateq((unsigned char*)SIGN,buf,6))
				auth=true;
			else
				throw error("Using incorrect protocol");
			hlpr->suck();
			fldb[0]=SERV_FLOOD;
			for(short i=1;i<SERV_FLOOD_SZ;i++)
				fldb[i]=0;
			for(short i=0;i<80;i++)
				hlpr->send(fldb,SERV_FLOOD_SZ);
			ashp=ship::pickspawn();
			for(short i=0;i<CLIENTS;i++)
				if(players[i] && players[i]->mshp==ashp)
					ashp=NULL;
			if(ashp)
			{
				sprintf(nam,"Player");
				assign(ashp);
			}
			else
				changecmod(CMOD_NAME);
		}
	}
	if(tcks%24==0)
	{
		sbnd=hlpr->getcount();
		if(sbnd>((cbnd*2)/5))
			urat+=6;
		if(sbnd<((cbnd*2)/5))
			urat-=2;
		if(urat<10)
			urat=10;
		if(urat>70)
			urat=70;
	}
	tout++;
	if(tout>250)
		throw error("Client timed out");
}

void server::action(short typ,short opr)
{
	switch(typ)
	{
		case CLIENT_ACCEL:
		if(mshp)
		{
			mshp->stat=ship::STAT_HOLD;
			if(opr==-2)
				mshp->accel(-1,true);	
			if(opr==-1)
				mshp->accel(-1,false);	
			if(opr==1)	
				mshp->accel(+1,false);	
			if(opr==2)	
				mshp->accel(+1,true);	
		}
		break;
		
		case CLIENT_TURN:
		if(mshp)
		{
			mshp->stat=ship::STAT_HOLD;
			if(opr==-1)
				mshp->turn(-1);
			if(opr==+1)
				mshp->turn(+1);
		}
		break;

		case CLIENT_SHOOT:
		if(mshp)
		{
			if(opr==0)
				mshp->shoot(false);
			if(opr==1)
				mshp->shoot(true);
		}
		break;

		case CLIENT_TRG:
		if(mshp)
		{
			mshp->stat=ship::STAT_HOLD;
			mshp->enem=NULL;
			mshp->frnd=NULL;
			mshp->plnt=NULL;
			if(opr>=0 && opr<(planet::ISIZE+ship::ISIZE))
			{
				if(opr<planet::ISIZE)
					mshp->plnt=planet::get(opr);
				else
				{
					mshp->enem=ship::get(opr-planet::ISIZE);
					if(mshp->enem==mshp)
						mshp->enem=NULL;
					if(mshp->enem)
						if(mshp->all->opposes(mshp->enem->all))
							registersound(mshp,SND_PROXIMITY);
						else
							registersound(mshp,SND_BEEP2);
				}
			}
			if(cmod==CMOD_SCAN || cmod==CMOD_HAIL || cmod==CMOD_WHOIS)
				changecmod(cmod);
			if(cmod==CMOD_REFIT)
				changecmod(CMOD_HAIL);
		}
		break;

		case CLIENT_CMOD:
		if(mshp)
		{
			if(opr==REQ_STAT)
				changecmod(CMOD_STAT);
			if(opr==REQ_EQUIP)
				changecmod(CMOD_EQUIP);
			if(opr==REQ_SCAN)
				changecmod(CMOD_SCAN);
			if(opr==REQ_HAIL)
				changecmod(CMOD_HAIL);
			if(opr==REQ_CHAT)
				changecmod(CMOD_CHAT);
			if(opr==REQ_WHOIS)
				changecmod(CMOD_WHOIS);
		}
		break;

		case CLIENT_CONS:
		cons(opr);
		break;

		case CLIENT_CHAR:
		if(opr==0)
		{
			input();
			inpb[0]='\0';
		}
		if(opr>=32 && opr<127)
		{
			if(opr>=97 && opr<=122)
				opr-=32;
			for(short i=0;i<64;i++)
			{
				if(inpb[i]=='\0')
				{
					inpb[i]=opr;
					inpb[i+1]='\0';
					break;
				}
			}
		}
		break;

		case CLIENT_BANDWIDTH:
		if(opr>cbnd)
			cbnd=opr;
		tout=0;
		break;

		default:
		break;
	}
}

void server::changecmod(short opr)
{
	char* txtp;
	char txt[1024]; //Text buffer for constructing output
	alliance* tali; //Alliance for possible choosing
	short spr; //Sprite to spritetocons with

	cmod=opr;
	txtp=txt;
	spr=-1;
	txt[0]='\0';
	if(mshp)
		registersound(mshp,SND_BEEP1);
	if(cmod!=CMOD_NAME && cmod!=CMOD_CHOOSE && !mshp)
		return;
	switch(opr)
	{
		case CMOD_NAME:
		printtocons("Choose player name\n");
		requestline(false);
		break;

		case CMOD_CHOOSE:
		txtp+=sprintf(txtp,"Choose alliance\n");
		for(short i=0;i<alliance::LIBSIZE;i++)
		{
			tali=alliance::get(i);
			if(tali)
				txtp+=sprintf(txtp,"[%hd] %s\n",i,tali->nam);
		}
		printtocons(txt);
		break;

		case CMOD_STAT:
		spr=mshp->serverreq(txt,CMOD_STAT,-1,mshp);
		printtocons(txt);
		spritetocons(spr);
		break;
		case CMOD_EQUIP:
		mshp->serverreq(txt,CMOD_EQUIP,-1,mshp);
		printtocons(txt);
		break;

		case CMOD_SCAN:
		if(mshp->enem)
		{
			spr=mshp->enem->serverreq(txt,CMOD_SCAN,-1,mshp);
			printtocons(txt);
			spritetocons(spr);
		}
		else
		{
			if(mshp->plnt)
			{
				spr=mshp->plnt->serverreq(txt,CMOD_SCAN,-1,mshp);
				printtocons(txt);
				spritetocons(spr);
			}
			else
			{
				printtocons("No target");
			}
		}
		break;

		case CMOD_HAIL:
		if(mshp->enem)
		{
			mshp->enem->serverreq(txt,CMOD_HAIL,-1,mshp);
			printtocons(txt);
		}
		else
		{
			if(mshp->plnt)
			{
				mshp->plnt->serverreq(txt,CMOD_HAIL,-1,mshp);
				printtocons(txt);
			}
			else
			{
				printtocons("No target");
			}
		}
		break;
		
		case CMOD_REFIT:
		if(mshp->plnt)
		{
			mshp->plnt->serverreq(txt,CMOD_REFIT,-1,mshp);
			printtocons(txt);
		}
		break;

		case CMOD_CHAT:
		if(mshp->enem && mshp->enem->ply)
		{
			printtocons("Send message to %s",playername(mshp->enem));
		}
		else
		{
			printtocons("Broadcast message");
		}
		requestline(false);
		break;

		case CMOD_WHOIS:
		txtp+=sprintf(txtp,"[Whois]\n\n");
		if(mshp->enem)
		{
			spr=mshp->enem->serverreq(txtp,CMOD_WHOIS,-1,mshp);
			txtp=txt+strlen(txt);
		}
		else
			txtp+=sprintf(txtp,"No target\n");
		txtp+=sprintf(txtp,"\n[1] Cycle to next player");
		printtocons(txt);
		spritetocons(spr);
		break;

		
	}
}

void server::cons(short opr)
{
	char txt[1024]; //Text buffer for constructing output
	alliance* tali; //Alliance to choose
	ship* tshp; //Temporary ship scratchpad

	if(opr<0)
		return;
	if(cmod!=CMOD_CHOOSE && !mshp)
		return;
	switch(cmod)
	{
		case CMOD_CHOOSE:
		tali=alliance::get(opr);
		if(tali)
		{
			tshp=ship::pickspawn(tali);
			if(tshp)
			{
				assign(tshp);
			}
			else
				printtomesg("No available spawn points");
		}
		else
		{
			printtomesg("No such alliance");
			changecmod(CMOD_CHOOSE);
		}
		break;

		case CMOD_EQUIP:
		mshp->serverreq(txt,CMOD_EQUIP,opr,mshp);
		break;

		case CMOD_SCAN:
		if(mshp->enem)
			mshp->enem->serverreq(txt,CMOD_SCAN,opr,mshp);
		else
			if(mshp->plnt)
				mshp->plnt->serverreq(txt,CMOD_SCAN,opr,mshp);
		break;

		case CMOD_HAIL:
		if(mshp->enem)
		{
			mshp->enem->serverreq(txt,CMOD_HAIL,opr,mshp);
			printtomesg(txt);
			changecmod(CMOD_HAIL);
		}
		else
		{
			if(mshp->plnt)
			{
				if(opr==4)
				{
					changecmod(CMOD_REFIT);
				}
				else
				{
					mshp->plnt->serverreq(txt,CMOD_HAIL,opr,mshp);
					printtomesg(txt);
					changecmod(CMOD_HAIL);
				}
			}
		}
		break;

		case CMOD_REFIT:
		if(mshp->plnt)
		{
			mshp->plnt->serverreq(txt,CMOD_REFIT,opr,mshp);
			printtomesg(txt);
			changecmod(CMOD_REFIT);
		}
		break;

		case CMOD_WHOIS:
		if(opr==1)
		{
			for(short i=0,j=(mshp->enem ? mshp->enem->self+1 : 0);i<ship::ISIZE;i++,j++)
			{
				if(j>=ship::ISIZE)
					j=0;
				tshp=ship::get(j);
				if(tshp && tshp->ply && tshp!=mshp)
				{
					mshp->enem=tshp;
					break;
				}
			}
			changecmod(CMOD_WHOIS);
		}
		break;
	}
}

void server::requestline(bool hide)
{
	unsigned char buf[SERV_READLN_SZ]; //For sending the request byte

	buf[0]=SERV_READLN;
	buf[1]=(unsigned char)hide;
	hlpr->send(buf,SERV_READLN_SZ);
}

void server::input()
{
	switch(cmod)
	{
		case CMOD_NAME:
		inpb[32]='\0';
		if(strlen(inpb)<2)
		{
			printtomesg("Name too short");
			inpb[0]='\0';
		}
		for(short i=0;i<CLIENTS;i++)
		{
			if(players[i] && players[i]!=this)
			{
				if(inpb[0]==players[i]->nam[0] && inpb[1]==players[i]->nam[1])
				{
					printtomesg("Name clashes with connected player \"%s\"",players[i]->nam);
					inpb[0]='\0';
					break;
				}
			}
		}
		if(inpb[0]!='\0')
		{
			sprintf(nam,"%s",inpb);
			log("Player %hd took name '%s'",self,nam);
			changecmod(CMOD_CHOOSE);
		}
		else
			changecmod(CMOD_NAME);
		break;

		case CMOD_CHAT:
		if(mshp)
		{
			if(mshp->enem && mshp->enem->ply)
			{
				hail(mshp,mshp->enem,inpb);
			}
			else
			{
				for(short i=0;i<CLIENTS;i++)
					if(players[i] && players[i]->mshp)
						hail(mshp,players[i]->mshp,inpb);
			}
			changecmod(CMOD_CHAT);
		}
		break;
	}
}

void server::printtocons(char* fmt,...)
{
	unsigned char buf[1028]; //Outgoing buffer
	va_list fmts;

	va_start(fmts,fmt);
	vsprintf((char*)buf+3,fmt,fmts);
	va_end(fmts);

	buf[0]=SERV_CONS;
	calc::shorttodat(strlen((char*)buf+3),buf+1);

	hlpr->send(buf,strlen((char*)buf+3)+3);
}

void server::spritetocons(short indx)
{
	unsigned char buf[SERV_CSPR_SZ]; //Outgoing buffer

	if(indx>=0)
	{
		buf[0]=SERV_CSPR;
		calc::shorttodat(indx,buf+1);
		hlpr->send(buf,SERV_CSPR_SZ);
	}
}

void server::printtomesg(char* fmt,...)
{
	unsigned char buf[68]; //Outgoing buffer
	va_list fmts;

	if(fmt[0]!='\0')
	{
		va_start(fmts,fmt);
		vsprintf((char*)buf+3,fmt,fmts);
		va_end(fmts);

		buf[0]=SERV_MESG;
		calc::shorttodat(strlen((char*)buf+3),buf+1);

		hlpr->send(buf,strlen((char*)buf+3)+3);
	}
}

void server::assign(ship* tshp)
{
	mshp=tshp;
	mshp->assign();
	log("Player %hd assigned to ship %d : %s",self,tshp->self,tshp->cls);
	changecmod(CMOD_STAT);
}

void server::uploads()
{
	unsigned char buf[SERV_SELF_SZ]; //Buffer for outgoing data

	if(mshp && tcks%(urat/10)==0)
	{
		mshp->netout(SERV_SELF,buf);
		hlpr->send(buf,SERV_SELF_SZ);
		calc::shorttodat(foc,buf+21);
		if(tcks%((urat*4)/10)==0)
			uploadplanets();
		uploadships();
		uploadfrags();
		hlpr->pump();
	}
}

void server::uploadplanets()
{
	unsigned char buf[256]; //Outgoing scratchpad buffer to use
	planet* tpln; //Concerned planet

	for(short i=0;i<planet::ISIZE;i++)
	{
		tpln=planet::get(i);
		if(tpln)
		{
			if(mshp->see(tpln))
			{
				if(!plnu[i])
				{
					tpln->netout(SERV_NEW,buf);
					hlpr->send(buf,SERV_NEW_SZ);
					tpln->netout(SERV_NAME,buf);
					hlpr->send(buf,SERV_NAME_SZ);
					plnu[i]=true;
				}
				tpln->netout(SERV_UPD,buf);
				hlpr->send(buf,SERV_UPD_SZ);
			}
			else
			{
				if(plnu[i])
				{
					buf[0]=SERV_DEL;
					calc::shorttodat(planet2pres(i),buf+1);
					hlpr->send(buf,SERV_DEL_SZ);
					plnu[i]=false;
				}
			}
		}
		else
		{
			if(plnu[i])
			{
				buf[0]=SERV_DEL;
				calc::shorttodat(planet2pres(i),buf+1);
				hlpr->send(buf,SERV_DEL_SZ);
				plnu[i]=false;
			}
		}
	}
}

void server::uploadships()
{
	unsigned char buf[256]; //Outgoing scratchpad buffer to use
	ship* tshp; //Concerned ship

	for(short i=0;i<ship::ISIZE;i++)
	{
		tshp=ship::get(i);
		if(tshp)
		{
			if(mshp->see(tshp))
			{
				if(!shpu[i])
				{
					tshp->netout(SERV_NEW,buf);
					hlpr->send(buf,SERV_NEW_SZ);
					tshp->netout(SERV_NAME,buf);
					hlpr->send(buf,SERV_NAME_SZ);
					shpu[i]=true;
				}
				tshp->netout(SERV_UPD,buf);
				if(mshp->all->opposes(tshp->all))
					buf[21]=1;
				hlpr->send(buf,SERV_UPD_SZ);
			}
			else
			{
				if(shpu[i])
				{
					buf[0]=SERV_DEL;
					calc::shorttodat(ship2pres(i),buf+1);
					hlpr->send(buf,SERV_DEL_SZ);
					shpu[i]=false;
				}
			}
		}
		else
		{
			if(shpu[i])
			{
				buf[0]=SERV_DEL;
				calc::shorttodat(ship2pres(i),buf+1);
				hlpr->send(buf,SERV_DEL_SZ);
				shpu[i]=false;
			}
		}
	}
}

void server::uploadfrags()
{
	unsigned char buf[256]; //Outgoing scratchpad buffer to use
	frag* tfrg; //Concerned frag

	for(short i=0;i<frag::ISIZE;i++)
	{
		tfrg=frag::get(i);
		if(tfrg)
		{
			if(mshp->see(tfrg))
			{
				if(frgu[i])
				{
					tfrg->netout(SERV_UPD,buf);
					hlpr->send(buf,SERV_UPD_SZ);
				}
				else
				{
					tfrg->netout(SERV_NEW,buf);
					hlpr->send(buf,SERV_NEW_SZ);
					tfrg->netout(SERV_UPD,buf);
					hlpr->send(buf,SERV_UPD_SZ);
					frgu[i]=true;
				}
			}
			else
			{
				if(frgu[i])
				{
					buf[0]=SERV_DEL;
					calc::shorttodat(frag2pres(i),buf+1);
					hlpr->send(buf,SERV_DEL_SZ);
					frgu[i]=false;
				}
			}
		}
		else
		{
			if(frgu[i])
			{
				buf[0]=SERV_DEL;
				calc::shorttodat(frag2pres(i),buf+1);
				hlpr->send(buf,SERV_DEL_SZ);	
				frgu[i]=false;
			}
		}
	}
}

void server::kill()
{
	unsigned char buf[SERV_DEL_SZ]; //Buffer for deleting player's ship clientside

	if(mshp)
	{
		uploadplanets();
		uploadships();
		uploadfrags();
		buf[0]=SERV_DEL;
		calc::shorttodat(ship2pres(mshp->self),buf+1);
		hlpr->send(buf,SERV_DEL_SZ);
		printtomesg("You have been destroyed: Game Over");
		log("Player %hd was destroyed",self);
		mshp=NULL;
	}
	changecmod(CMOD_CHOOSE);
}

void server::hilight(ship* tshp)
{
	unsigned char buf[SERV_HILIGHT_SZ]; //Buffer for sending hilight information

	if(tshp)
	{
		buf[0]=SERV_HILIGHT;
		calc::shorttodat(ship2pres(tshp->self),buf+1);
		hlpr->send(buf,SERV_HILIGHT_SZ);
	}
}

server* server::players[CLIENTS];
long server::tcks;
TCPsocket server::lstn;
