/*
	ship.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

#include <stdio.h>
#include "calc.h"
#include "error.h"
#include "frag.h"
#include "database.h"
#include "planet.h"
#include "equip.h"
#include "server.h"
#include "alliance.h"
#include "constants.h"
#include "protocol.h"

struct slot //Internal equipment slot of ship
{
	pol pos; //Position (ang=-1 indicates not external)
	short face; //Facing direction
	equip* item; //Equipment item index
	short rdy; //Readiness
	short cap; //Capacity
};

class ship //Spaceship type in game
{
	public:
	const static short ISIZE=128; //Size of ship index
	const static short LIBSIZE=16; //Size of ship type library
	enum {STAT_HOLD=0,STAT_FOLLOW=1,STAT_ATTACK=2,STAT_PLANET=4,STAT_EVADE=8,STAT_CRIPPLED=16}; //Ship status type
	enum {PRCH_FUEL,PRCH_ARMS,PRCH_HULL}; //Codes for different purchases

	ship(cord loc,short typ,alliance* tali); //Spawns a new ship at given place, with given type and alliance
	~ship(); //Destructor, deletes ship resolving all dependencies (like frags and other ships)

	static void init(); //Initialise all ship datastructures
	static void loadlib(); //Load the library of ship types
	static void purgeall(); //Clean up the ship index (erasure)
	static void simulateall(); //Simulate the physics of all the ships
	static void behaveall(); //Simulates primitive behaviour for all ships
	static void saveall(); //Save all ships to database
	static void loadall(); //Load all ships from database
	static ship* get(short indx); //Returns a pointer to a ship of given index
	static ship* pickspawn(alliance* tali); //Returns a ship suitable for spawning into from given alliance
	static ship* pickspawn(); //Pick a ship with player attribute set (for restoring the player in a local game)
	static void makefresh(); //Sets player attributes as false on all ships, so there's no game restore effect (use for headless server)

	void turn(short dir); //Turns the ship by specified amount(in 1/100ths of an angle)
	void accel(short dir,bool wrp); //Accelerates the ship positively or negatively (+1 or -1 in dir), wrp deciding if you are willing to make the impulse<=>warp transition
	void shoot(bool torp); //Shoot at target with torpedoes if torp, otherwise phasers
	bool see(ship* tshp); //Returns true if the given ship is visible to the current one, otherwise false
	bool see(planet* tpln); //Returns true if the given planet is visible to the current ship, otherwise false
	bool see(frag* tfrg); //Returns true if the given frag is visible to the current ship, otherwise false
	short serverreq(char* txt,short cmod,short opr,ship* mshp); //Handles a server request for information/action from this ship, with the given comm mode, operand and player's ship, writing the text into txt and returning the sprite index (-1 if n/a)
	short freemass(); //Returns free mass on board
	void cloak(); //Cloak the ship
	void uncloak(); //Uncloak the ship
	void shieldsup(); //Raise the shields
	void shieldsdown(); //Drop the shields
	void netout(short typ,unsigned char* buf); //Get type of data from ship into a network buffer
	bool colldetect(cord frgl,vect frgv); //Given an intruder location and vector, determines if a collision occurs, returning true if it does, false otherwise
	void hit(short mag,cord frgl,vect frgv,ship* src); //Damages the ship with given magnitude, also location and vector of striking object, plus the source of the attack (expected to be frag owner)
	void assign(); //Sets the ship up for entrance of a player
	long purchase(short prch,short ripo,bool buy);
	long purchase(equip* prch,short ripo,bool buy); //Returns the cost of a given purchase type at given rip-off index, with buy set actually deducts the cash
	bool transport(planet* to,bool beam);
	bool transport(ship* to,bool beam); //Returns if possible to transport to the given destination, with beam set it will actually carry it out
	void credit(long amnt); //Credit the ship's cash by given amount, and inform them

	short self; //Self index in the ship database
	bool ply; //Is a player?
	alliance* all; //Alliance
	cord loc; //Co-ordinates 
	char cls[65]; //Ship class 
	short spr; //Sprite index
	ship* frnd; //Friend target
	ship* enem; //Enemy target
	planet* plnt; //Planetary target
	short stat; //Behaviour status

	private:
	ship(); //Thinned down constructor, used only for making slots to load the ship library into
	ship(short self); //Constructor that loads given self index from database
	void physics(); //Handles physics of motion on this ship
	void autonav(); //Executes auto-pilot considerations on ship
	void save(); //Saves the ship data to database
	void load(); //Load ship data from database, object should already have been selected
	void loadlink(); //Extension to load that only resolves the links between ships
	void maintain(); //Does routine work such as recharging shields and power, reloading weapons
	void behave(); //Execute characteristic behaviour for this ship
	void act(short actn); //Perform given action type, called by behave()
	void lynch(); //Get all ships in the area to come attack this ship, usually as a result of friendly fire
	void firecontrol(); //Makes AI decisions concerning phaser fire

	static ship* lib[LIBSIZE]; //Ship type database
	static ship* ships[ISIZE]; //Main ship list
	static short mstr,istr; //Master strobe and individual strobe, used for periodic behaviours

	short typ; //Ship type index
	short w[36],h[36]; //Width and height (halved)
	short sspr; //Shield sprite
	short fspr; //Frag sprite
	short fsnd; //Frag sound
	short dsnd; //Destruction sound
	pol vel; //Velocity
	short trn; //Turning speed
	long mip; //Maximum impulse velocity (<100)
	long aip; //Impulse acceleration
	long mwp; //Maximum warp velocity
	long awp; //Warp acceleration
	long srng; //Sensor range
	short mss; //Mass
	short hul; //Hull integrity
	short mhul; //Maximum hull integrity
	short pow; //Current power level
	short rpow; //Power regeneration rate
	short mpow; //Maximum power level
	short shd; //Shield level
	short rshd; //Shield regeneration rate
	short mshd; //Maximum shield level
	short shon; //Shields on?
	short clon; //Cloak on?
	long ful; //Fuel level
	long mful; //Maximum fuel level
	slot slots[32]; //Equipment
	short mlck; //Mass locked?
	long cash; //Cash onboard
	vect mov; //Caches velocity vector for use by other functions
};

ship::ship(cord loc,short typ,alliance* tali)
{
	if(!tali)
		throw error("No alliance given for this ship");
	self=-1;
	if(!lib[typ])
		throw error("No such ship type");
	*this=*lib[typ];
	for(short i=0;i<ISIZE && self==-1;i++)
	{
		if(!ships[i])
		{
			self=i;
			ships[self]=this;
		}
	}
	if(self==-1)
		throw error("No free slots in the ship index");
	this->typ=typ;
	this->loc=loc;
	this->all=tali;

	shd=0;
	pow=mpow;
	ful=mful;
	hul=mhul;
	ply=false;
	vel.ang=calc::rnd(360);
	vel.rad=0;
	mov.xx=0;
	mov.yy=0;

	frnd=NULL;
	enem=NULL;
	plnt=NULL;
	stat=STAT_HOLD;
	shon=1;
	clon=0;
	mlck=true;
	cash=0;
	for(short i=0;i<32;i++)
		slots[i].rdy=0;
}

ship::~ship()
{
	if(self>=0 && self<ISIZE)
		ships[self]=NULL;
	for(short i=0;i<ISIZE;i++)
	{
		if(ships[i])
		{
			if(ships[i]->enem==this)
				ships[i]->enem=NULL;
			if(ships[i]->frnd==this)
				ships[i]->frnd=NULL;
		}
	}
	frag::notifydeath(this);
	server::notifydeath(this);
}

void ship::init()
{
	for(short i=0;i<ISIZE;i++)
		ships[i]=NULL;
	for(short i=0;i<LIBSIZE;i++)
		lib[i]=NULL;
}

void ship::loadlib()
{
	char obsc[16]; //Object name scratchpad

	for(short i=0;i<LIBSIZE;i++)
	{
		if(!lib[i])
		{
			try
			{
				lib[i]=new ship();
				sprintf(obsc,"ShipLib%hd",i);
				database::switchobj(obsc);
				lib[i]->load();
			}
			catch(error it)
			{
				lib[i]=NULL;
			}
		}
	}
}

void ship::purgeall()
{
	for(short i=0;i<ISIZE;i++)
		if(ships[i])
			delete ships[i];
}

void ship::simulateall()
{
	for(short i=0;i<ISIZE;i++)
	{
		if(ships[i])
		{
			ships[i]->physics();
			ships[i]->maintain();
		}
	}
}

void ship::behaveall()
{
	mstr=(mstr+1)%1000;
	for(short i=0;i<ISIZE;i++)
	{
		if(ships[i] && (ships[i]->stat&STAT_CRIPPLED)!=STAT_CRIPPLED)
		{
			istr=(mstr+i*7)%400;
			if(ships[i]->stat!=STAT_HOLD)
				ships[i]->autonav();
			if(!ships[i]->ply)
				ships[i]->firecontrol();
		}	
	}
	for(short i=0,j;i<10;i++)
	{
		j=calc::rnd(ISIZE);
		if(ships[j] && (ships[j]->stat&STAT_CRIPPLED)!=STAT_CRIPPLED)
		{
			if(!(ships[j]->ply))
				ships[j]->behave();
			ships[j]->mlck=planet::masslock(ships[j]->loc);
		}
	}
}

void ship::saveall()
{
	char obsc[16]; //Object name scratchpad

	for(short i=0;i<ISIZE;i++)
	{
		if(ships[i])
		{
			sprintf(obsc,"Ship%hd",i);
			database::putobject(obsc);
			ships[i]->save();
		}
	}
}

void ship::loadall()
{
	char obsc[16]; //Object name scratchpad

	for(short i=0;i<ISIZE;i++)
	{
		try
		{
			new ship(i);
		}
		catch(error it)
		{
		}
	}
	for(short i=0;i<ISIZE;i++)
	{
		if(ships[i])
		{
			sprintf(obsc,"Ship%hd",i);
			database::switchobj(obsc);
			ships[i]->loadlink();
		}
	}
}

ship* ship::get(short indx)
{
	if(indx>=0 && indx<ISIZE)
	{
		if(ships[indx])
			return ships[indx];
		else
			return NULL;
	}
	else
	{
		return NULL;
	}
}

ship* ship::pickspawn(alliance* tali)
{
	short styp; //Type to look for

	styp=tali->sinv[0];
	for(short i=0;i<ISIZE;i++)
		if(ships[i] && ships[i]->typ==styp && !(ships[i]->ply))
			return ships[i];
	return NULL;
}

ship* ship::pickspawn()
{
	for(short i=0;i<ISIZE;i++)
		if(ships[i] && ships[i]->ply)
			return ships[i];
	return NULL;
}

void ship::makefresh()
{
	for(short i=0;i<ISIZE;i++)
		if(ships[i] && ships[i]->ply)
			ships[i]->ply=false;
}

void ship::turn(short dir)
{
	if(pow>0)
	{
		if(dir==+1)
			vel.ang+=trn;
		if(dir==-1)
			vel.ang-=trn;
		vel.ang=vel.ang%360;
		if(vel.ang<0)
			vel.ang+=360;
	}
}

void ship::accel(short dir,bool wrp)
{
	short pcon; //Power consumption
	long nsp; //New speed
	short astr; //Accleration strobe (for dealing with sub-integer acceleration
	
	nsp=vel.rad;
	if(vel.rad<=mip)
	{
		astr=mstr%10;
		if(dir==+1)
		{
			nsp=vel.rad+aip/10;
			if((aip-aip/10)<=astr)
				nsp+=1;
		}
		if(dir==-1)
		{
			nsp=vel.rad-aip/10;
			if((aip-aip/10)<=astr)
				nsp-=1;
		}
		if(nsp>mip)
			if(wrp)
				nsp=100;
			else
				nsp=mip;
	}
	if(vel.rad>=100)
	{
		if(dir==+1)
			nsp=vel.rad+awp;
		if(dir==-1)
			nsp=vel.rad-awp;
		if(nsp<100)
			if(wrp)
				nsp=mip;
			else
				nsp=100;
	}

	if(nsp>mip && nsp<100)
		nsp=mip;
	if(nsp>mwp)
		nsp=mwp;
	if(nsp<0)
		nsp=0;

	if(nsp>=100)
		pcon=(((vel.rad-nsp)*mss)/2000);
	else
		pcon=(((vel.rad-nsp)*mss)/2);
	if(pcon<0)
		pcon=-pcon;
	if((nsp>=100 && vel.rad<100) || (nsp<100 && vel.rad>=100))
		pcon=0;

	if(pow>=pcon)
	{
		pow-=pcon;
		vel.rad=nsp;
	}
}

void ship::shoot(bool torp)
{
	pol ptmp;
	vect vtmp; //Temporaries for calculations
	pol ptrg; //Polar to target
	vect vtrg; //Vector to target
	cord cemt; //Location to emit frag at
	pol pemt; //Polar velocity to emit frag at
	vect vemt; //Vector to emit frag at
	short dif; //Angle difference
	long rng; //Range
	bool can; //Can shoot from this slot?
	equip* lnch; //Equipment doing the launching

	if(!see(enem))
		return;
	if(vel.rad>=100)
		return;
	rng=0;
	vtrg.xx=enem->loc.x-loc.x;
	vtrg.yy=enem->loc.y-loc.y;
	ptrg=calc::vecttopol(vtrg);
	for(short i=0;i<32;i++)
	{
		if(slots[i].item && ((torp && slots[i].item->typ==equip::LAUNCHER) || (!torp && slots[i].item->typ==equip::PHASER)))
		{
			lnch=slots[i].item;
			can=false;
			if(lnch->typ==equip::PHASER)
				rng=lnch->rng*lnch->trck;
			if(lnch->typ==equip::LAUNCHER)
				rng=(srng*10)/9;
			if(ptrg.rad<=rng)
				can=true;
			if(can)
			{
				dif=ptrg.ang-(vel.ang+slots[i].face);
				if(dif>180)
					dif=dif-360;
				if(dif<-180)
					dif=dif+360;
				if(dif>(lnch->acov) || dif<-(lnch->acov))
					can=false;	
				if(dif>(lnch->acov) || dif<-(lnch->acov))
					can=false;		
				if(lnch->typ==equip::PHASER && pow<lnch->pow)
					can=false;
				if(lnch->typ==equip::LAUNCHER && !slots[i].cap>0)
					can=false;
				if(slots[i].rdy!=0)
					can=false;
			}
			if(can)
			{
				uncloak();
				
				ptmp=slots[i].pos;
				ptmp.ang=(ptmp.ang+vel.ang)%360;
				vtmp=calc::poltovect(ptmp);
				
				cemt.x=vtmp.xx+loc.x;
				cemt.y=vtmp.yy+loc.y;

				pemt=ptrg;
				pemt.rad=lnch->trck;

				vemt=calc::poltovect(pemt);
				vemt.xx+=mov.xx;
				vemt.yy+=mov.yy;

				if(lnch->typ==equip::PHASER)
				{
					pow-=lnch->pow;
					new frag(cemt,frag::ENERGY,lnch->spr,lnch->col,enem,this,vemt,((ptrg.ang+5)/10)%36,lnch->pow,0,lnch->rng);
				}
				if(lnch->typ==equip::LAUNCHER)
				{
					slots[i].cap--;
					new frag(cemt,frag::HOMER,lnch->spr,lnch->col,enem,this,vemt,0,lnch->pow,lnch->trck,lnch->rng);
				}
				server::registernoise(this,lnch->snd);
				slots[i].rdy=lnch->rdy;
				if(torp)
					break;
			}
		}
	}
}

bool ship::see(ship* tshp)
{
	if(!tshp)
		return false;
	if(tshp==this)
		return true;
	if(tshp->clon==100)
		return false;
	if((tshp->loc.x-loc.x)>srng)
		return false;
	if((tshp->loc.x-loc.x)<-srng)	
		return false;
	if((tshp->loc.y-loc.y)>srng)
		return false;
	if((tshp->loc.y-loc.y)<-srng)	
		return false;
	return true;		
}

bool ship::see(planet* tpln)
{
	if(!tpln)
		return false;
	if(tpln->typ==planet::STAR)
		return true;
	if((tpln->loc.x-loc.x)>srng)
		return false;
	if((tpln->loc.x-loc.x)<-srng)
		return false;
	if((tpln->loc.y-loc.y)>srng)
		return false;
	if((tpln->loc.y-loc.y)<-srng)
		return false;
	return true;		
}

bool ship::see(frag* tfrg)
{
	if(!tfrg)
		return false;
	if((tfrg->loc.x-loc.x)>srng)
		return false;
	if((tfrg->loc.x-loc.x)<-srng)
		return false;
	if((tfrg->loc.y-loc.y)>srng)
		return false;
	if((tfrg->loc.y-loc.y)<-srng)
		return false;
	return true;		
}

short ship::serverreq(char* txt,short cmod,short opr,ship* mshp)
{
	bool cdev; //Cloaking device?
	char spd[32]; //Speed
	char* pnam; //Player name

	if(!mshp)
		return -1;
	switch(cmod)
	{
		case CMOD_STAT:
		case CMOD_SCAN:
		if(opr==-1)
		{
			if(mshp->see(this))
			{
				txt+=sprintf(txt,"%s\n",cls);
				if(mshp->all->opposes(all))
					txt+=sprintf(txt,"Alignment:%s [hostile]\n",all->nam);
				else
					txt+=sprintf(txt,"Alignment:%s\n",all->nam);
				pnam=server::playername(this);
				if(pnam)
					txt+=sprintf(txt,"Commanded by %s\n",pnam);
				if(shon)
					txt+=sprintf(txt,"\nShields: Raised\n");
				else
					txt+=sprintf(txt,"\nShields: Down\n");
				calc::getspeed(mwp,spd);
				txt+=sprintf(txt,"Maximum velocity: %s\n",spd);
				txt+=sprintf(txt,"Shield capability: %hd\n",mshd);
				txt+=sprintf(txt,"Maximum power capacity: %hd\n",mpow);
				txt+=sprintf(txt,"Fuel storage: %ld/%ld\n",ful,mful);
				if(this==mshp)
					txt+=sprintf(txt,"\nCredits: %ld\nAvailable mass: %hd\n",cash,freemass());
				if(this==mshp->enem)
					txt+=sprintf(txt,"\n[1] Lay in an intercept course\n");
				return spr;
			}
			else
			{
				txt+=sprintf(txt,"Target not visible\n");
				if(this==mshp->enem)
					txt+=sprintf(txt,"\n[1] Lay in an intercept course\n");
				return -1;
			}
		}
		if(opr==1 && this==mshp->enem)
		{
			mshp->stat=STAT_ATTACK;
		}
		break;

		
		case CMOD_EQUIP:
		if(opr==-1)
		{
			cdev=false;
			txt+=sprintf(txt,"Internal systems\n\n");
			for(short i=0;i<32;i++)
			{
				if(slots[i].item)
				{
					if(slots[i].item->typ==equip::CLOAK)
						cdev=true;
					if(slots[i].cap>=0)
						txt+=sprintf(txt,"  %s [%hd]\n",slots[i].item->nam,slots[i].cap);
					else
						txt+=sprintf(txt,"  %s\n",slots[i].item->nam);
				}
				else
				{
					if(slots[i].pos.rad>=0)
					{
						if(slots[i].face<=90 || slots[i].face>=270)
							txt+=sprintf(txt," <Free forward port>\n");
						else
							txt+=sprintf(txt," <Free rear port>\n");
					}
				}
			}
			txt+=sprintf(txt,"\n");
			txt+=sprintf(txt,"[1] Toggle shields\n");
			if(cdev)
				txt+=sprintf(txt,"[2] Toggle cloak\n");
		}
		if(opr==1)
			if(shon)
				shieldsdown();
			else
				shieldsup();
		if(opr==2)
			if(clon==0)
				cloak();
			else
				uncloak();
		break;

		case CMOD_HAIL:
		if((stat&STAT_CRIPPLED)==STAT_CRIPPLED)
		{
			if(opr==-1)
			{
				txt+=sprintf(txt,"Hailing %s\n\n",cls);
				txt+=sprintf(txt,"Vessel is disabled\n\n[1] Attempt to recover it");
			}
			if(opr==1)
			{
				if(mshp->transport(this,true))
				{
					enem=NULL;
					plnt=NULL;
					frnd=mshp;
					stat=STAT_FOLLOW;
					this->all=mshp->all;
					for(short i=0;i<ISIZE;i++)
						if(ships[i] && ships[i]->enem==this)
							ships[i]->enem=NULL;
					txt+=sprintf(txt,"Vessel successfully appropriated");
				}
				else
					txt+=sprintf(txt,"Cannot establish transporter lock");
			}
		}
		else
		{
			if(frnd==mshp)
			{
				if(opr==-1)
				{
					txt+=sprintf(txt,"Hailing %s\n\n",cls);
					txt+=sprintf(txt,"Vessel is under your command\n\n[1] Transfer to this vessel");
				}
				if(opr==1)
				{
					if(mshp->transport(this,true))
					{
						server::transfer(mshp,this);
						txt+=sprintf(txt,"Transfer of command successful");
					}
					else
					{
						if(transport(mshp,true))
						{
							server::transfer(mshp,this);
							txt+=sprintf(txt,"Transfer of command successful (had to use reverse transport)");
						}
						else
							txt+=sprintf(txt,"Cannot establish transporter lock");
					}
				}
			}
			else
			{
				txt+=sprintf(txt,"Hailing %s\n\n",cls);
				txt+=sprintf(txt,"No reply");
			}
		}
		break;
		
		case CMOD_WHOIS:
		if(ply)
		{
			txt+=sprintf(txt,"Player: %s\n",server::playername(this));
			txt+=sprintf(txt,"Alliance: %s\n",all->nam);
			return spr;
		}
		else
			txt+=sprintf(txt,"Target not player controlled\n");
		break;
	}
	return -1;
}

short ship::freemass()
{
	short out; //Outputted free space

	out=mss;
	for(short i=0;i<32;i++)
	{
		if(slots[i].item)
			out-=slots[i].item->mss;
	}
	return out;
}

void ship::cloak()
{
	slot swap; //For swapping the cloaking device into convenient position

	if(clon==0)
	{
		for(short i=0;i<32;i++)
		{
			if(slots[i].item && slots[i].item->typ==equip::CLOAK)
			{
				swap=slots[0];
				slots[0]=slots[i];
				slots[i]=swap;
				clon=+1;
				server::registernoise(this,slots[0].item->snd);
				break;
			}
		}
	}
}

void ship::uncloak()
{
	if(clon>0)
	{
		clon=-clon;
		server::registernoise(this,slots[0].item->snd);
	}
}

void ship::shieldsup()
{
	if(mshd>0)
	shon=1;
}

void ship::shieldsdown()
{
	shon=0;
}

void ship::netout(short typ,unsigned char* buf)
{
	buf[0]=typ;
	buf+=1;

	calc::shorttodat(ship2pres(self),buf);
	buf+=2;
	switch(typ)
	{
		case SERV_SELF:
		if(mss>0)
			calc::shorttodat((100*hul)/mhul,buf);
		else
			calc::shorttodat(0,buf);
		buf+=2;

		if(mpow>0)
			calc::shorttodat((100*pow)/mpow,buf);
		else
			calc::shorttodat(0,buf);
		buf+=2;

		if(mshd>0)
			calc::shorttodat((100*shd)/mshd,buf);
		else
			calc::shorttodat(0,buf);
		buf+=2;

		if(mful>0)
			calc::shorttodat((100*ful)/mful,buf);
		else
			calc::shorttodat(0,buf);
		buf+=2;
		calc::longtodat(srng,buf);
		buf+=4;
		calc::longtodat(LIMIT,buf);
		buf+=4;
		if(plnt)
		{
			calc::shorttodat(planet2pres(plnt->self),buf);
		}
		else
		{
			if(enem)
				calc::shorttodat(ship2pres(enem->self),buf);
			else
				calc::shorttodat(-1,buf);
		}
		buf+=2;
		calc::shorttodat(-1,buf);
		buf+=2;
		calc::shorttodat(-1,buf);
		break;

		case SERV_NEW:
		*buf=PT_SHIP;
		buf+=1;
		calc::shorttodat(spr,buf);
		buf+=2;
		calc::shorttodat(-1,buf);
		buf+=2;
		break;

		case SERV_NAME:
		sprintf((char*)buf,"%s",cls);
		buf+=64;
		sprintf((char*)buf,"%s",all->nam);
		buf+=64;
		break;

		case SERV_UPD:
		calc::longtodat(loc.x,buf);
		buf+=4;
		calc::longtodat(loc.y,buf);
		buf+=4;
		calc::longtodat(mov.xx,buf);
		buf+=4;
		calc::longtodat(mov.yy,buf);
		buf+=4;
		calc::shorttodat(vel.ang,buf);
		buf+=2;
		*buf=0;
		buf+=1;
		if(clon>=0)
			*buf=100-clon;
		else
			*buf=100+clon;
		buf+=1;
		break;

		default:
		break;
	}
}

bool ship::colldetect(cord frgl,vect frgv)
{
	short rot; //Target rotation
	long x1,y1,x2,y2,xx,yy; //Target bounding box

	rot=((vel.ang+5)/10)%36;
	xx=(frgv.xx-mov.xx)/2;
	yy=(frgv.yy-mov.yy)/2;
	if(xx<0)
		xx=-xx;
	if(yy<0)
		yy=-yy;
	x1=loc.x-(w[rot]*3)/2-xx;
	y1=loc.y-(h[rot]*3)/2-yy;
	x2=loc.x+(w[rot]*3)/2+xx;
	y2=loc.y+(h[rot]*3)/2+yy;
	if(frgl.x>x1 && frgl.x<x2 && frgl.y>y1 && frgl.y<y2)
		return true;
	else
		return false;
}

void ship::hit(short mag,cord frgl,vect frgv,ship* src)
{
	cord tmpc;
	vect tmpv; //Temporary for working out detonations
	short rot; //Target rotation
	short ndeb; //Number of debris bits

	uncloak();
	rot=((vel.ang+5)/10)%36;
	if(rot<0)
		rot+=36;
	shd-=mag;
	server::registershake(this,mag/100);
	if(src && enem!=src && !(all->opposes(src->all)) && src->ply)
		src->lynch();
	if(shd>0)
	{
		new frag(frgl,frag::DEBRIS,sspr,-1,NULL,this,mov,calc::rnd(36),0,0,2);
	}
	else
	{
		frgl.x=(frgl.x+loc.x)/2;
		frgl.y=(frgl.y+loc.y)/2;
		for(short i=0;i<5;i++)
		{
			tmpv=mov;
			tmpv.xx+=calc::rnd(mag/60)-calc::rnd(mag/60);
			tmpv.yy+=calc::rnd(mag/60)-calc::rnd(mag/60);
			while(tmpv.xx>6)
				tmpv.xx/=2;
			while(tmpv.yy>6)
				tmpv.yy/=2;
			frgl.x+=calc::rnd(2)-calc::rnd(2);
			frgl.y+=calc::rnd(2)-calc::rnd(2);
			new frag(frgl,frag::DEBRIS,frag::FIRE,-1,NULL,this,tmpv,calc::rnd(36),0,0,calc::rnd(5)+5);
		}
		shd=0;
		server::registernoise(this,fsnd);
		if(hul<mhul/2 && pow<mpow/3 && !ply)
			stat=STAT_CRIPPLED;
	}
	hul-=(mag*4)/(shd+10);
	if(hul<0)
	{
		hul=0;
		ndeb=mss/8+4;
		if(ndeb>70)
			ndeb=70;
		for(short i=0;i<ndeb;i++)
		{	
			tmpc.x=calc::rnd(2*w[rot])-calc::rnd(2*w[rot]);
			tmpc.x+=loc.x;
			tmpc.y=calc::rnd(2*h[rot])-calc::rnd(2*h[rot]);
			tmpc.y+=loc.y;
			tmpv=mov;
			tmpv.xx+=calc::rnd(3+5*mag/mss)-calc::rnd(3+5*mag/mss);
			tmpv.yy+=calc::rnd(3+5*mag/mss)-calc::rnd(3+5*mag/mss);
			if(calc::rnd(10)<3)
				new frag(tmpc,frag::DEBRIS,frag::FIRE,-1,NULL,this,tmpv,calc::rnd(36),0,0,calc::rnd(20)+5);
			else
				new frag(tmpc,frag::DEBRIS,fspr,-1,NULL,this,tmpv,calc::rnd(36),0,0,calc::rnd(20)+5);
		}
		if(src && src->all->opposes(all))
		{
			server::hail(NULL,src,"Target destroyed; bounty paid");
			src->credit(mss*3);
		}
		server::registernoise(this,dsnd);
		delete this;
	}
}

void ship::assign()
{
	ply=true;
	enem=NULL;
	frnd=NULL;
	if(plnt)
		stat=STAT_PLANET;
	else
		stat=STAT_HOLD;
}

long ship::purchase(short prch,short ripo,bool buy)
{
	long cost; //Value to output

	cost=0;
	if(prch==PRCH_FUEL)
	{
		if(ful<mful)
			cost=ripo/4;
		if(buy)
			ful=mful;
	}
	if(prch==PRCH_ARMS)
	{
		for(short i=0;i<32;i++)
		{
			if(slots[i].item && slots[i].item->typ==equip::LAUNCHER)
			{
				if(slots[i].cap<slots[i].item->cap)
				{
					cost=(slots[i].item->cost*ripo)/1000;
					if(buy)
						slots[i].cap=slots[i].item->cap;
					break;
				}
			}
		}
	}
	if(prch==PRCH_HULL)
	{
		cost=((mhul-hul)*ripo)/100;
		if(buy)
			hul=mhul;
	}
	if(buy && cash>=cost)
		cash-=cost;
	if(cost>cash)
		cost=0;
	return cost;
}

long ship::purchase(equip* prch,short ripo,bool buy)
{
	long cost; //Value to output

	if(!prch)
		return 0;
	cost=0;
	if(!(prch->mss>freemass()))
	{
		for(short i=0;i<32;i++)
		{
			if(!slots[i].item)
			{
				if((slots[i].pos.rad>=0 && (prch->typ==equip::PHASER || prch->typ==equip::LAUNCHER)) || slots[i].pos.rad==-1)
				{
					cost=(prch->cost*ripo)/100;
					if(buy)
					{
						slots[i].item=prch;
						slots[i].cap=prch->cap;
						slots[i].rdy=prch->rdy;
					}
					break;
				}
			}
		}
	}
	if(buy && cash>=cost)
		cash-=cost;
	if(cost>cash)
		cost=0;
	return cost;
}

bool ship::transport(planet* to,bool beam)
{
	vect vto;
	pol pto; //Vectors to the target

	vto.xx=to->loc.x-loc.x;
	vto.yy=to->loc.y-loc.y;
	pto=calc::vecttopol(vto);
	if(shon)
		return false;
	if(clon!=0)
		return false;
	for(short i=0;i<32;i++)
	{
		if(slots[i].item && slots[i].item->typ==equip::TRANSPORTER && slots[i].rdy==0 && pow>=slots[i].item->pow && slots[i].item->rng>=pto.rad)
		{
			if(beam)
			{
				pow-=slots[i].item->pow;
				slots[i].rdy=slots[i].item->rdy;
				server::registersound(this,slots[i].item->snd);
			}
			return true;
		}
	}
	return false;
}

bool ship::transport(ship* to,bool beam)
{
	vect vto;
	pol pto; //Vectors to the target

	vto.xx=to->loc.x-loc.x;
	vto.yy=to->loc.y-loc.y;
	pto=calc::vecttopol(vto);
	if(shon)
		return false;
	if(clon!=0)
		return false;
	if(to->shon || to->clon!=0)
		return false;
	for(short i=0;i<32;i++)
	{
		if(slots[i].item && slots[i].item->typ==equip::TRANSPORTER && slots[i].rdy==0 && pow>=slots[i].item->pow && slots[i].item->rng>=pto.rad)
		{
			if(beam)
			{
				pow-=slots[i].item->pow;
				slots[i].rdy=slots[i].item->rdy;
				server::registersound(this,slots[i].item->snd);
			}
			return true;
		}
	}
	return false;
}

void ship::credit(long amnt)
{
	char txt[17]; //Description of crediting

	cash+=amnt;
	if(cash>999999999)
		cash=999999999;
	sprintf(txt,"Credited %ld C",amnt);
	server::hail(NULL,this,txt);
}

ship::ship()
{
	self=-1;	
}

ship::ship(short self)
{
	char obsc[16]; //Object name scratchpad

	this->self=-1;
	if(!(self>=0 && self<ISIZE))
		throw error("Given ship index for loading out of range");
	if(ships[self])
		throw error("Given ship index slot already in use");
	sprintf(obsc,"Ship%hd",self);
	database::switchobj(obsc);
	load();
	if(!all)
		throw error("No alliance found for this ship");
	this->self=self;
	ships[self]=this;
}

void ship::physics()
{
	if(vel.rad>=100 && mlck)
		vel.rad=mip;
	if(vel.rad<100 && vel.rad>mip)
		vel.rad=mip;

	if(loc.x>LIMIT || loc.x<-LIMIT || loc.y>LIMIT || loc.y<-LIMIT)
		vel.ang=(vel.ang+180)%360;
	if(loc.x>LIMIT)
		loc.x=LIMIT;
	if(loc.x<-LIMIT)
		loc.x=-LIMIT;
	if(loc.y>LIMIT)
		loc.y=LIMIT;
	if(loc.y<-LIMIT)
		loc.y=-LIMIT;
	
	mov=calc::poltovect(vel);

	loc.x+=mov.xx;
	loc.y+=mov.yy;
}

void ship::autonav()
{
	vect vtrg; //Vector to target
	pol ptrg; //Polar to target
	short dd; //Directional difference
	short tol; //Angular tolerance
	bool erat; //eratic approach?
	ship* tshp; //Target ship
	planet* tpln; //Target planet

	tshp=NULL;
	tpln=NULL;
	if(plnt && (stat&STAT_PLANET)==STAT_PLANET)
	{
		tpln=plnt;
		tshp=NULL;
	}
	if(enem && (stat&STAT_ATTACK)==STAT_ATTACK)
	{
		tshp=enem;
		tpln=NULL;
	}
	if(frnd && (stat&STAT_FOLLOW)==STAT_FOLLOW)
	{
		tshp=frnd;
		tpln=NULL;
	}
	if((stat&STAT_EVADE)==STAT_EVADE)
		erat=true;
	else
		erat=false;
	if(!(tshp==NULL && tpln==NULL))
	{
		if(tpln)
		{
			if(erat)
			{
				vtrg.xx=(self*497)%(800+vel.rad)-(400+vel.rad/2);
				vtrg.yy=(self*273)%(800+vel.rad)-(400+vel.rad/2);
			}
			else
			{
				vtrg.xx=(self*497)%800-400;
				vtrg.yy=(self*273)%800-400;
			}
		}
		if(tshp)
		{
			if(erat)
			{
				if(istr>(200+calc::rnd(100)-50))
					ptrg.ang=tshp->vel.ang+45+(istr-self*29)%270;
				else
					ptrg.ang=tshp->vel.ang+45-(istr+self*29)%270;
				ptrg.rad=100+(self*17)%((istr+srng)/16);
			}
			else
			{
				ptrg.ang=tshp->vel.ang+90+(self*29)%180;
				ptrg.rad=100+(self*17)%(srng/16);
			}
			vtrg=calc::poltovect(ptrg);
		}
		
		if(tpln==NULL)
		{
			vtrg.xx+=tshp->loc.x-loc.x;
			vtrg.yy+=tshp->loc.y-loc.y;
		}
		else
		{
			vtrg.xx+=tpln->loc.x-loc.x;
			vtrg.yy+=tpln->loc.y-loc.y;
		}

		ptrg=calc::vecttopol(vtrg);

		dd=ptrg.ang-vel.ang;
		if(dd>180)
			dd=dd-360;
		if(dd<-180)
			dd=dd+360;

		if(tshp && !see(tshp))
			ptrg.rad-=srng/3;
		tol=trn*2+2;
		if(vel.rad<=5) 
			tol=90;

		ptrg.rad-=150;
		if(erat && tshp)
		{
			ptrg.rad+=srng/4;
			tol+=45;
		}

		if(ptrg.rad<0)
			dd=0;

		if(dd<tol && dd>-tol)
			dd=0;
		if(dd==0 && ptrg.rad>0)
		{
			if(vel.rad>=100)
			{
				if(tpln)
					ptrg.rad-=5000+vel.rad*3;
				if(ptrg.rad && (ptrg.rad/vel.rad) && ((vel.rad)/(2*ptrg.rad/vel.rad))>=awp-30)
					accel(-1,true);
				else
					accel(+1,true);
			}
			else
			{
				if(ptrg.rad && ((5*vel.rad*vel.rad)/(ptrg.rad))>=aip)
					accel(-1,true);
				else
					accel(+1,true);
			}
		}
		else
			accel(-1,true);
		if(dd>0)
			turn(+1);
		if(dd<0)
			turn(-1);
	}
}

void ship::save()
{
	char atsc[33]; //Attribute scratchpad

	database::putvalue("Type",typ);
	database::putvalue("Class",cls);
	database::putvalue("ShipSprite",spr);
	database::putvalue("Width",w[0]);
	database::putvalue("Height",h[0]);
	if(sspr)
		database::putvalue("ShieldSprite",sspr);
	if(fspr)
		database::putvalue("FragSprite",fspr);
	if(fsnd)
		database::putvalue("FragSound",fsnd);
	if(dsnd)
		database::putvalue("DeathSound",dsnd);
	if(all)
		database::putvalue("Team",all->self);
	database::putvalue("Status",stat);
	database::putvalue("XLoc",loc.x);
	database::putvalue("YLoc",loc.y);
	database::putvalue("Heading",vel.ang);
	database::putvalue("Speed",vel.rad);
	database::putvalue("TurnRate",trn);
	database::putvalue("SublightLimit",mip);
	database::putvalue("SublightAcceleration",aip);
	database::putvalue("WarpLimit",mwp);
	database::putvalue("WarpAcceleration",awp);
	database::putvalue("SensorRange",srng);
	database::putvalue("Mass",mss);
	database::putvalue("HullStrength",hul);
	database::putvalue("HullStrengthLimit",mhul);
	database::putvalue("Power",pow);
	database::putvalue("PowerRecharge",rpow);
	database::putvalue("PowerLimit",mpow);
	database::putvalue("Shield",shd);
	database::putvalue("ShieldRecharge",rshd);
	database::putvalue("ShieldLimit",mshd);
	database::putvalue("Fuel",ful);
	database::putvalue("FuelLimit",mful);
	database::putvalue("ShieldStatus",shon);
	database::putvalue("CloakStatus",clon);
	if(frnd)
		database::putvalue("FriendTarget",frnd->self);
	if(enem)
		database::putvalue("EnemyTarget",enem->self);
	if(plnt)
		database::putvalue("PlanetTarget",plnt->self);
	database::putvalue("MassLock",mlck);
	database::putvalue("Cash",cash);
	for(short i=0;i<32;i++)
	{
		if(slots[i].item || slots[i].pos.rad!=-1)
		{
			sprintf(atsc,"Slot%hdAngle",i);
			database::putvalue(atsc,slots[i].pos.ang);
			sprintf(atsc,"Slot%hdRadius",i);
			database::putvalue(atsc,slots[i].pos.rad);
			sprintf(atsc,"Slot%hdFace",i);
			database::putvalue(atsc,slots[i].face);
			sprintf(atsc,"Slot%hdItem",i);
			if(slots[i].item)
				database::putvalue(atsc,slots[i].item->self);
			else
				database::putvalue(atsc,-1);
			sprintf(atsc,"Slot%hdReadiness",i);
			database::putvalue(atsc,slots[i].rdy);
			sprintf(atsc,"Slot%hdCapacity",i);
			database::putvalue(atsc,slots[i].cap);
		}
	}
	if(ply)
		database::putvalue("IsPlayer",(long)1);
	else
		database::putvalue("IsPlayer",(long)0);
}

void ship::load()
{
	char atsc[33]; //Attribute scratchpad
	pol bpol;
	vect vct1,vct2; //Temporaries for calculating the bounding box

	typ=database::getvalue("Type");
	database::getvalue("Class",cls);
	spr=database::getvalue("ShipSprite");
	w[0]=database::getvalue("Width");
	h[0]=database::getvalue("Height");
	for(short i=1;i<36;i++)
	{
		bpol.ang=i*10;
		bpol.rad=h[0];
		vct1=calc::poltovect(bpol);
		bpol.ang=(i*10+90)%360;
		bpol.rad=w[0];
		vct2=calc::poltovect(bpol);
		if(vct1.xx<0)
			vct1.xx=-vct1.xx;
		if(vct2.xx<0)
			vct2.xx=-vct2.xx;
		if(vct1.yy<0)
			vct1.yy=-vct1.yy;
		if(vct2.yy<0)
			vct2.yy=-vct2.yy;
		if(vct1.xx>vct2.xx)
			w[i]=(short)vct1.xx;
		else
			w[i]=(short)vct2.xx;
		if(vct1.yy>vct2.yy)
			h[i]=(short)vct1.yy;
		else
			h[i]=(short)vct2.yy;
	}
	sspr=database::getvalue("ShieldSprite");
	fspr=database::getvalue("FragSprite");
	fsnd=database::getvalue("FragSound");
	dsnd=database::getvalue("DeathSound");
	all=alliance::get(database::getvalue("Team"));
	stat=database::getvalue("Status");
	loc.x=database::getvalue("XLoc");
	loc.y=database::getvalue("YLoc");
	vel.ang=database::getvalue("Heading");
	vel.rad=database::getvalue("Speed");
	trn=database::getvalue("TurnRate");
	mip=database::getvalue("SublightLimit");
	aip=database::getvalue("SublightAcceleration");
	mwp=database::getvalue("WarpLimit");
	awp=database::getvalue("WarpAcceleration");
	srng=database::getvalue("SensorRange");
	mss=database::getvalue("Mass");
	hul=database::getvalue("HullStrength");
	mhul=database::getvalue("HullStrengthLimit");
	pow=database::getvalue("Power");
	rpow=database::getvalue("PowerRecharge");
	mpow=database::getvalue("PowerLimit");
	shd=database::getvalue("Shield");
	rshd=database::getvalue("ShieldRecharge");
	mshd=database::getvalue("ShieldLimit");
	ful=database::getvalue("Fuel");
	mful=database::getvalue("FuelLimit");
	shon=database::getvalue("ShieldStatus");
	clon=database::getvalue("CloakStatus");
	plnt=planet::get(database::getvalue("PlanetTarget"));
	
	mlck=database::getvalue("MassLock");
	cash=database::getvalue("Cash");
	for(short i=0;i<32;i++)
	{
		sprintf(atsc,"Slot%hdAngle",i);
		slots[i].pos.ang=database::getvalue(atsc);
		sprintf(atsc,"Slot%hdRadius",i);
		slots[i].pos.rad=database::getvalue(atsc);
		sprintf(atsc,"Slot%hdFace",i);
		slots[i].face=database::getvalue(atsc);
		if(slots[i].face==-1)
			slots[i].face=slots[i].pos.ang;
		sprintf(atsc,"Slot%hdItem",i);
		slots[i].item=equip::get(database::getvalue(atsc));
		sprintf(atsc,"Slot%hdReadiness",i);
		slots[i].rdy=database::getvalue(atsc);
		sprintf(atsc,"Slot%hdCapacity",i);
		slots[i].cap=database::getvalue(atsc);
	}
	mov.xx=0;
	mov.yy=0;
	if(database::getvalue("IsPlayer")==1)
		ply=true;
	else
		ply=false;
}

void ship::loadlink()
{
	frnd=get(database::getvalue("FriendTarget"));
	enem=get(database::getvalue("EnemyTarget"));
}

void ship::maintain()
{
	if((stat&STAT_CRIPPLED)==STAT_CRIPPLED)
	{
		shon=0;
		clon=0;
	}
	else
	{
		if(pow<mpow && ful>0)
		{
			ful-=rpow;
			pow+=rpow;
			if(ful<0)
				ful=0;
		}
		if(pow>mpow)
		{
			ful+=pow-mpow;
			pow=mpow;
		}
	}
	
	for(short i=0;i<32;i++)
	{
		if(slots[i].rdy>0)
			slots[i].rdy--;
	}
	if(shon==1 && pow>=rshd)
	{
		shd+=rshd;
		pow-=rshd;
		if(shd>mshd)
			shd=mshd;
	}
	else
	{
		shd/=2;
	}
	if(clon>0)
	{
		if(slots[0].item && slots[0].item->typ==equip::CLOAK)
		{
			if(pow>=(slots[0].item->pow*mss)/20)
				pow-=(slots[0].item->pow*mss)/20;
			else
				uncloak();
			clon+=2;
			if(clon>100)
				clon=100;
		}
		else
			uncloak();
	}
	if(clon<0)
		clon++;

	if(pow<0)
		pow=0;
	if(ful==0 && !ply && calc::rnd(100)==0)
		delete this;
}

void ship::behave()
{
	for(short i=0;i<32;i++)
		if(all->bhvs[i]>0)
			if(calc::rnd(100)<=all->bhvs[i])
				act(i);

	if(plnt && clon==0 && vel.rad>=100 && all->opposes(plnt->all))
		cloak();
	if(frnd && clon==0 && frnd->clon!=0)
		cloak();
	if(vel.rad==0 && clon>0)
		uncloak();

	if(frnd && frnd->frnd)
		frnd=frnd->frnd;
	if(frnd==this)
		frnd=NULL;
	if(frnd && !see(frnd))
	{
		stat=STAT_FOLLOW;
		enem=NULL;
		plnt=NULL;
	}

	if(enem || frnd)
		plnt=NULL;

	if(enem)
		shieldsup();
	else
		shieldsdown();
}

void ship::act(short actn)
{
	switch(actn)
	{
		case alliance::COWARD:
		if(enem && enem->mpow>mpow && enem->mshd>mshd)
		{
			enem=NULL;
			if(frnd)
				stat=STAT_EVADE|STAT_FOLLOW;
			else
				plnt=planet::pick(all);
			if(plnt)
				stat=STAT_PLANET|STAT_EVADE;
			else
				stat=STAT_HOLD;
		}
		break;

		case alliance::ATTACK:
		if((stat&STAT_EVADE)!=STAT_EVADE)
		{
			for(short i=0,j=0;i<200;i++)
			{
				j=calc::rnd(ISIZE);
				if(ships[j] && see(ships[j]) && all->opposes(ships[j]->all) && (ships[j]->stat&STAT_CRIPPLED)!=STAT_CRIPPLED)
				{
					enem=ships[j];
					if(!frnd)
						stat=STAT_ATTACK;
				}
			}
		}
		break;

		case alliance::MAKEFRIEND:
		if(!frnd && !enem)
		{
			for(short i=0,j=0;i<200;i++)
			{
				j=calc::rnd(ISIZE);
				if(ships[j] && ships[j]!=this && see(ships[j]) && all==ships[j]->all && !ships[j]->frnd && ships[j]!=enem && !ships[j]->ply && (ships[j]->stat&STAT_CRIPPLED)!=STAT_CRIPPLED)
				{
					frnd=ships[j];
					plnt=NULL;
					stat=STAT_FOLLOW;
				}
			}
		}
		break;
	
		case alliance::DESERT:
		if(frnd || enem)
		{
			frnd=NULL;
			enem=NULL;
			plnt=planet::pick(all);
			if(plnt)
				stat=STAT_PLANET|STAT_EVADE;
			else
				stat=STAT_HOLD;
		}
		break;

		case alliance::TRAVELHOME:
		if(self%2 && !enem && !frnd && ((plnt && see(plnt)) || !plnt))
		{
			plnt=planet::pick(all);
			stat=STAT_PLANET;
		}
		break;

		case alliance::TRAVELFRIEND:
		if(self%2 && !enem && !frnd && ((plnt && see(plnt)) || !plnt))
		{
			plnt=planet::pickally(all);
			stat=STAT_PLANET;
		}
		break;

		case alliance::TRAVELENEMY:
		if(self%2 && !enem && !frnd && ((plnt && see(plnt)) || !plnt))
		{
			plnt=planet::pickhostile(all);
			stat=STAT_PLANET;
		}
		break;

		case alliance::TORPEDO:
		if(enem && enem->shd*3>shd)
			shoot(true);
		break;
		
		case alliance::ATTACKWILD:
		if(enem)
			stat=STAT_ATTACK|STAT_EVADE;
		break;
		
		case alliance::ATTACKSTRAIGHT:
		stat=stat&~STAT_EVADE;
		break;

		case alliance::ATTACKFORMATION:
		if(frnd)
		{
			stat=STAT_FOLLOW;
			if(frnd->enem && all->opposes(frnd->enem->all))
				enem=frnd->enem;
		}
		default:
		break;
	}
}

void ship::lynch()
{
	for(short i=0;i<ISIZE;i++)
	{
		if(ships[i] && !ships[i]->ply && ships[i]!=this && !ships[i]->enem && ships[i]->see(this))
		{
			ships[i]->enem=this;
			ships[i]->frnd=NULL;
			ships[i]->stat=STAT_ATTACK;
		}
	}
}

void ship::firecontrol()
{
	short dd; //For comparing direction

	if(enem && see(enem))
	{
		dd=vel.ang-enem->vel.ang;
		if(dd<0)
			dd=-dd;
		if(dd>180)
			dd=360-dd;
		if(dd<=30 || dd>=150 || vel.rad<=5 || enem->vel.rad<=5)
			shoot(false);
	}
}

ship* ship::lib[LIBSIZE];
ship* ship::ships[ISIZE];
short ship::mstr;
short ship::istr;
