/*
	ship.h
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/

struct pol;
struct vect;
struct cord;
class alliance;
class planet;
class equip;
class frag;

struct slot //Internal equipment slot of ship
{
	pol pos; //Position (ang=-1 indicates not external)
	short face; //Facing direction
	equip* item; //Equipment item index
	short rdy; //Readiness
	short cap; //Capacity
};

class ship //Spaceship type in game
{
	public:
	const static short ISIZE=128; //Size of ship index
	const static short LIBSIZE=16; //Size of ship type library
	enum {STAT_HOLD=0,STAT_FOLLOW=1,STAT_ATTACK=2,STAT_PLANET=4,STAT_EVADE=8,STAT_CRIPPLED=16}; //Ship status type
	enum {PRCH_FUEL,PRCH_ARMS,PRCH_HULL}; //Codes for different purchases

	ship(cord loc,short typ,alliance* tali); //Spawns a new ship at given place, with given type and alliance
	~ship(); //Destructor, deletes ship resolving all dependencies (like frags and other ships)

	static void init(); //Initialise all ship datastructures
	static void loadlib(); //Load the library of ship types
	static void purgeall(); //Clean up the ship index (erasure)
	static void simulateall(); //Simulate the physics of all the ships
	static void behaveall(); //Simulates primitive behaviour for all ships
	static void saveall(); //Save all ships to database
	static void loadall(); //Load all ships from database
	static ship* get(short indx); //Returns a pointer to a ship of given index
	static ship* pickspawn(alliance* tali); //Returns a ship suitable for spawning into from given alliance
	static ship* pickspawn(); //Pick a ship with player attribute set (for restoring the player in a local game)
	static void makefresh(); //Sets player attributes as false on all ships, so there's no game restore effect (use for headless server)

	void turn(short dir); //Turns the ship by specified amount(in 1/100ths of an angle)
	void accel(short dir,bool wrp); //Accelerates the ship positively or negatively (+1 or -1 in dir), wrp deciding if you are willing to make the impulse<=>warp transition
	void shoot(bool torp); //Shoot at target with torpedoes if torp, otherwise phasers
	bool see(ship* tshp); //Returns true if the given ship is visible to the current one, otherwise false
	bool see(planet* tpln); //Returns true if the given planet is visible to the current ship, otherwise false
	bool see(frag* tfrg); //Returns true if the given frag is visible to the current ship, otherwise false
	short serverreq(char* txt,short cmod,short opr,ship* mshp); //Handles a server request for information/action from this ship, with the given comm mode, operand and player's ship, writing the text into txt and returning the sprite index (-1 if n/a)
	short freemass(); //Returns free mass on board
	void cloak(); //Cloak the ship
	void uncloak(); //Uncloak the ship
	void shieldsup(); //Raise the shields
	void shieldsdown(); //Drop the shields
	void netout(short typ,unsigned char* buf); //Get type of data from ship into a network buffer
	bool colldetect(cord frgl,vect frgv); //Given an intruder location and vector, determines if a collision occurs, returning true if it does, false otherwise
	void hit(short mag,cord frgl,vect frgv,ship* src); //Damages the ship with given magnitude, also location and vector of striking object, plus the source of the attack (expected to be frag owner)
	void assign(); //Sets the ship up for entrance of a player
	long purchase(short prch,short ripo,bool buy);
	long purchase(equip* prch,short ripo,bool buy); //Returns the cost of a given purchase type at given rip-off index, with buy set actually deducts the cash
	bool transport(planet* to,bool beam);
	bool transport(ship* to,bool beam); //Returns if possible to transport to the given destination, with beam set it will actually carry it out

	short self; //Self index in the ship database
	bool ply; //Is a player?
	alliance* all; //Alliance
	cord loc; //Co-ordinates 
	char cls[65]; //Ship class 
	short spr; //Sprite index
	ship* frnd; //Friend target
	ship* enem; //Enemy target
	planet* plnt; //Planetary target
	short stat; //Behaviour status

	private:
	ship(); //Thinned down constructor, used only for making slots to load the ship library into
	ship(short self); //Constructor that loads given self index from database
	void physics(); //Handles physics of motion on this ship
	void autonav(); //Executes auto-pilot considerations on ship
	void save(); //Saves the ship data to database
	void load(); //Load ship data from database, object should already have been selected
	void loadlink(); //Extension to load that only resolves the links between ships
	void maintain(); //Does routine work such as recharging shields and power, reloading weapons
	void behave(); //Execute characteristic behaviour for this ship
	void act(short actn); //Perform given action type, called by behave()
	void lynch(); //Get all ships in the area to come attack this ship, usually as a result of friendly fire
	void firecontrol(); //Makes AI decisions concerning phaser fire

	static ship* lib[LIBSIZE]; //Ship type database
	static ship* ships[ISIZE]; //Main ship list
	static short mstr,istr; //Master strobe and individual strobe, used for periodic behaviours

	short typ; //Ship type index
	short w[36],h[36]; //Width and height (halved)
	short sspr; //Shield sprite
	short fspr; //Frag sprite
	short fsnd; //Frag sound
	short dsnd; //Destruction sound
	pol vel; //Velocity
	short trn; //Turning speed
	long mip; //Maximum impulse velocity (<100)
	long aip; //Impulse acceleration
	long mwp; //Maximum warp velocity
	long awp; //Warp acceleration
	long srng; //Sensor range
	short mss; //Mass
	short hul; //Hull integrity
	short mhul; //Maximum hull integrity
	short pow; //Current power level
	short rpow; //Power regeneration rate
	short mpow; //Maximum power level
	short shd; //Shield level
	short rshd; //Shield regeneration rate
	short mshd; //Maximum shield level
	short shon; //Shields on?
	short clon; //Cloak on?
	long ful; //Fuel level
	long mful; //Maximum fuel level
	slot slots[32]; //Equipment
	short mlck; //Mass locked?
	long cash; //Cash onboard
	vect mov; //Caches velocity vector for use by other functions
};


