/*
	sound.cc
	
	(c) Richard Thrippleton
	Licensing terms are in the 'LICENSE' file
	If that file is not included with this source then permission is not given to use this source in any way whatsoever.
*/


#include <SDL.h>
#include <string.h>
#include "error.h"

class sound;

struct channel //Used to represent a virtual channel for mixing
{
	sound* snd; //Sound currently playing
	long pntr; //Playing pointer for each sound in queue
	short div; //Div factor for each sound
};

class sound //A sound in the database
{
	public:
	static const short ISIZE=256; //Maximum index size of sounds

	static void init(); //Initialise the sound system datastructures
	static void start(); //Start the sound system
	static void stop(); //Stop the sound system
	static sound* get(short indx); //Returns sound of given index from the database

	void play(short div); //Schedules sound to play with given div factor

	private:
	sound(short self); //Constructor, give self-index of sound

	static void callback(void* null,Uint8* bfil,int lfil); //Callback slave for the SDL sound thread

	void load(); //Loads sound in from appropriate file

	static channel chns[16]; //Virtual channels
	static SDL_AudioSpec spec; //Obtained audio device specification
	static bool on; //Sound on?
	static sound* sounds[ISIZE]; //Playable sound index

	short self; //Self-index value of sound
	Uint8* buff; //Audio buffer
	Uint32 blen; //Buffer length
	bool imem; //In memory and loaded?
	bool miss; //Sound missing? Keep this as a record and try not to load it again
};

void sound::init()
{
	on=false;
	for(short i=0;i<ISIZE;i++)
		sounds[i]=NULL;
	for(short i=0;i<16;i++)
		chns[i].snd=NULL;
}

void sound::start()
{
	SDL_AudioSpec want; //Desired audio specification

	if(SDL_InitSubSystem(SDL_INIT_AUDIO)==-1)
		throw new error(SDL_GetError());
	if(on)
		return;
	want.freq=22050;
	want.format=AUDIO_S16SYS;
	want.channels=1;
	want.samples=1024;
	want.callback=sound::callback;
	if(SDL_OpenAudio(&want,&spec)==-1)
		on=false;
	else
	{
		on=true;
		SDL_PauseAudio(0);
	}
}

void sound::stop()
{
	if(!on)
		return;
	on=false;
	SDL_CloseAudio();
}

sound* sound::get(short indx)
{
	if(!(indx>=0 && indx<ISIZE))
		return NULL;
	if(!sounds[indx])
	{
		sounds[indx]=new sound(indx);
	}
	return sounds[indx];
}

void sound::play(short div)
{
	if(miss || !on)
		return;
	if(!imem)
	{
		try
		{
			load();
		}
		catch(error it)
		{
			miss=true;
		}
	}
	for(short i=0;i<16;i++)
		if(chns[i].snd==this && chns[i].pntr<64)
			return;
	for(short i=0;i<16;i++)
	{
		if(!chns[i].snd)
		{
			chns[i].snd=this;
			chns[i].pntr=0;
			chns[i].div=div;
			break;
		}	
	}
}

sound::sound(short self)
{
	this->self=self;
	miss=false;
	imem=false;
}

void sound::callback(void* null,Uint8* bfil,int lfil)
{
	long plen; //Length to blit to stream

	for(short i=0;i<16;i++)
	{
		if(chns[i].snd)
		{
			plen=chns[i].snd->blen-chns[i].pntr;
			if(plen>0)
			{
				if(plen>lfil)
					plen=lfil;
				SDL_MixAudio(bfil,&(chns[i].snd->buff[chns[i].pntr]),plen,SDL_MIX_MAXVOLUME/chns[i].div);
				chns[i].pntr+=plen;
			}
			else
			{
				chns[i].snd=NULL;
			}
		}
	}
}

void sound::load()
{
	SDL_AudioSpec wspc; //Wav specification
	SDL_AudioCVT cvt; //Wav to native format conversion structure
	char* path; //Path to load sprite from

	path=new char[strlen(DATADIR)+32];
	sprintf(path,"%s/snd/%i.wav",DATADIR,self);
	if(!SDL_LoadWAV(path,&wspc,&buff,&blen))
	{
		delete(path);
		throw error("Can't load wave file");
	}
	delete(path);
	if(SDL_BuildAudioCVT(&cvt,wspc.format,wspc.channels,wspc.freq,spec.format,spec.channels,spec.freq)==-1)
		throw error("Couldn't build the audio conversion structure");
	cvt.buf=new Uint8[blen*cvt.len_mult];
	cvt.len=blen;
	memcpy(cvt.buf,buff,cvt.len);
	SDL_ConvertAudio(&cvt);
	SDL_FreeWAV(buff);
	buff=cvt.buf;
	blen=cvt.len_cvt;
	imem=true;
}

channel sound::chns[16];
SDL_AudioSpec sound::spec;
bool sound::on;
sound* sound::sounds[ISIZE];
